# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for the Charmcraft-specific python plugin."""

import pathlib
import shlex

import pytest
import pytest_check

from charmcraft.parts import plugins


def test_get_build_environment(
    python_plugin: plugins.PythonPlugin, install_path: pathlib.Path
):
    env = python_plugin.get_build_environment()

    assert env["PIP_NO_BINARY"] == ":all:"


def test_get_venv_directory(
    python_plugin: plugins.PythonPlugin, install_path: pathlib.Path
):
    assert python_plugin._get_venv_directory() == install_path / "venv"


@pytest.mark.parametrize("constraints", [[], ["constraints.txt"]])
@pytest.mark.parametrize("requirements", [[], ["requirements.txt"]])
@pytest.mark.parametrize("packages", [[], ["distro==1.4.0"]])
def test_get_package_install_commands(
    tmp_path: pathlib.Path,
    python_plugin: plugins.PythonPlugin,
    build_path: pathlib.Path,
    install_path: pathlib.Path,
    constraints: list[str],
    requirements: list[str],
    packages: list[str],
    check,
):
    spec = {
        "plugin": "python",
        "source": str(tmp_path),
        "python-constraints": constraints,
        "python-requirements": requirements,
        "python-packages": packages,
    }
    python_plugin._options = plugins.PythonPluginProperties.unmarshal(spec)
    python_plugin._get_pip = lambda: "/python -m pip"
    copy_src_cmd = (
        f"cp --archive --recursive --reflink=auto {build_path}/src {install_path}"
    )
    copy_lib_cmd = (
        f"cp --archive --recursive --reflink=auto {build_path}/lib {install_path}"
    )

    actual = python_plugin._get_package_install_commands()

    with check():
        assert actual[0].startswith("/python -m pip")
    with check():
        assert actual[1].startswith("/python -m pip")
    split_install_command = shlex.split(actual[0])
    for constraints_file in constraints:
        pytest_check.is_in(f"--constraint={constraints_file}", split_install_command)
    for requirements_file in requirements:
        pytest_check.is_in(f"--requirement={requirements_file}", split_install_command)
    for package in packages:
        pytest_check.is_in(package, split_install_command)
    pytest_check.is_not_in(copy_src_cmd, actual)
    pytest_check.is_not_in(copy_lib_cmd, actual)

    (build_path / "src").mkdir()

    pytest_check.is_in(copy_src_cmd, python_plugin._get_package_install_commands())
    pytest_check.is_not_in(copy_lib_cmd, python_plugin._get_package_install_commands())

    (build_path / "lib").mkdir()

    pytest_check.is_in(copy_src_cmd, python_plugin._get_package_install_commands())
    pytest_check.is_in(copy_lib_cmd, python_plugin._get_package_install_commands())

    (build_path / "src").rmdir()

    pytest_check.is_not_in(copy_src_cmd, python_plugin._get_package_install_commands())
    pytest_check.is_in(copy_lib_cmd, python_plugin._get_package_install_commands())


def test_get_rm_command(
    python_plugin: plugins.PythonPlugin, install_path: pathlib.Path
):
    assert (
        f"rm -rf {install_path / 'venv/bin'}/!(activate)"
        in python_plugin.get_build_commands()
    )


def test_no_get_rm_command(
    tmp_path, python_plugin: plugins.PythonPlugin, install_path: pathlib.Path
):
    spec = {
        "plugin": "python",
        "source": str(tmp_path),
        "python-keep-bins": True,
    }
    python_plugin._options = plugins.PythonPluginProperties.unmarshal(spec)
    assert (
        f"rm -rf {install_path / 'venv/bin'}/!(activate)"
        not in python_plugin.get_build_commands()
    )



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Charmcraft-specific uv plugin."""

from pathlib import Path

from craft_parts.plugins import uv_plugin
from overrides import override

from charmcraft import utils


class UvPlugin(uv_plugin.UvPlugin):
    @override
    def get_build_environment(self) -> dict[str, str]:
        return utils.extend_python_build_environment(super().get_build_environment())

    @override
    def _get_venv_directory(self) -> Path:
        return self._part_info.part_install_dir / "venv"

    @override
    def _get_pip(self) -> str:
        return 'uv pip --python="${PARTS_PYTHON_VENV_INTERP_PATH}"'

    @override
    def _get_package_install_commands(self) -> list[str]:
        # Find the `uv sync` command and modify it to not install the project
        orig_cmds = super()._get_package_install_commands()
        for idx, cmd in enumerate(orig_cmds):
            if cmd.startswith("uv sync"):
                orig_cmds[idx] += " --no-install-project"
                break

        return [
            *orig_cmds,
            *utils.get_charm_copy_commands(
                self._part_info.part_build_dir, self._part_info.part_install_dir
            ),
        ]

    @override
    def _should_remove_symlinks(self) -> bool:
        return True

    @override
    def get_build_commands(self) -> list[str]:
        return [
            *super().get_build_commands(),
            *utils.get_venv_cleanup_commands(
                self._get_venv_directory(), keep_bins=False
            ),
        ]



from time import sleep

import flask

app = flask.Flask(__name__)


@app.route("/")
def index():
    return "Hello, world!\n"


@app.route("/io")
def pseudo_io():
    sleep(2)
    return "ok\n"

if __name__ == "__main__":
    app.run()



# This file is part of charmcraft.
#
# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import datetime
import os
import pathlib
import subprocess
import sys

import craft_parts_docs

import charmcraft

project_dir = pathlib.Path(__file__).parents[1].resolve()

project = "Charmcraft"
author = "Canonical"
release = charmcraft.__version__
if ".post" in release:
    # The commit hash in the dev release version confuses the spellchecker
    release = "dev"

copyright = "2023-%s, %s" % (datetime.date.today().year, author)

# region Configuration for canonical-sphinx
ogp_site_url = "https://documentation.ubuntu.com/charmcraft/"
ogp_site_name = project
ogp_image = "https://assets.ubuntu.com/v1/253da317-image-document-ubuntudocs.svg"

slug = "charmcraft"

html_context = {
    "product_page": "juju.is",
    "product_tag": "_static/assets/juju-logo-no-text.png",
    "github_url": "https://github.com/canonical/charmcraft",
    "github_issues": "https://github.com/canonical/charmcraft/issues",
    "discourse": "https://discourse.charmhub.io",
    "matrix": "https://matrix.to/#/#charmhub-charmcraft:ubuntu.com",
}

# Target repository for the edit button on pages
html_theme_options = {
    "source_edit_link": "https://github.com/canonical/charmcraft",
}

# Sitemap configuration: https://sphinx-sitemap.readthedocs.io/
html_baseurl = "https://documentation.ubuntu.com/charmcraft/"

if "READTHEDOCS_VERSION" in os.environ:
    version = os.environ["READTHEDOCS_VERSION"]
    sitemap_url_scheme = "{version}{link}"
else:
    sitemap_url_scheme = "latest/{link}"

# Template and asset locations
extensions = [
    "canonical_sphinx",
    "pydantic_kitbash",
    "sphinx_sitemap",
    "sphinx_substitution_extensions",
]

# Copy extra files to the _static dir during build
html_static_path = ["_static"]
templates_path = ["_templates"]

# Files for the cookie banner
html_css_files = [
    'css/cookie-banner.css'
]

html_js_files = [
    'js/bundle.js',
]

# endregion

# region General configuration
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions.extend(
    [
        "sphinx.ext.ifconfig",
        "sphinx.ext.intersphinx",
        "sphinx.ext.viewcode",
        "sphinx.ext.coverage",
        "sphinx.ext.doctest",
        "sphinx_toolbox",
        "sphinx_toolbox.more_autodoc",
        "sphinx.ext.autodoc",  # Must be loaded after more_autodoc
        "sphinxcontrib.details.directive",
        "sphinx_toolbox.collapse",
        "sphinxcontrib.details.directive",
        "sphinx.ext.napoleon",
        "sphinx_autodoc_typehints",  # must be loaded after napoleon
        "sphinxext.rediraffe",
    ]
)

# endregion

exclude_patterns = [
    # Workaround for https://github.com/canonical/pydantic-kitbash/issues/49
    "common/craft-parts/reference/part_properties.rst",

    "_build",
    "Thumbs.db",
    ".DS_Store",
    "env",
    "sphinx-starter-pack",
    # Excluded here because they are either included explicitly in other
    # documents (so they generate "duplicate label" errors) or they aren't
    # used in this documentation at all (so they generate "unreferenced"
    # errors).
    "common/craft-parts/explanation/lifecycle.rst",
    "common/craft-parts/explanation/overlay_parameters.rst",
    "common/craft-parts/explanation/overlays.rst",
    "common/craft-parts/explanation/parts.rst",
    "common/craft-parts/explanation/how_parts_are_built.rst",
    "common/craft-parts/explanation/dump_plugin.rst",
    "common/craft-parts/explanation/file-migration.rst",
    "common/craft-parts/explanation/gradle_plugin.rst",
    "common/craft-parts/explanation/overlay_step.rst",
    "common/craft-parts/how-to/craftctl.rst",
    "common/craft-parts/how-to/customise-the-build-with-craftctl.rst",
    "common/craft-parts/how-to/include_files.rst",
    "common/craft-parts/how-to/use_parts.rst",
    "common/craft-parts/how-to/override_build.rst",
    "common/craft-parts/reference/partition_specific_output_directory_variables.rst",
    "common/craft-parts/reference/step_output_directories.rst",
    "common/craft-parts/reference/plugins/ant_plugin.rst",
    "common/craft-parts/reference/plugins/autotools_plugin.rst",
    "common/craft-parts/reference/plugins/cargo_use_plugin.rst",
    "common/craft-parts/reference/plugins/cmake_plugin.rst",
    "common/craft-parts/reference/plugins/dotnet_plugin.rst",
    "common/craft-parts/reference/plugins/dotnet_v2_plugin.rst",
    "common/craft-parts/reference/plugins/go_plugin.rst",
    "common/craft-parts/reference/plugins/gradle_plugin.rst",
    "common/craft-parts/reference/plugins/jlink_plugin.rst",
    "common/craft-parts/reference/plugins/make_plugin.rst",
    "common/craft-parts/reference/plugins/maven_plugin.rst",
    "common/craft-parts/reference/plugins/maven_use_plugin.rst",
    "common/craft-parts/reference/plugins/meson_plugin.rst",
    "common/craft-parts/reference/plugins/npm_plugin.rst",
    "common/craft-parts/reference/plugins/poetry_plugin.rst",
    "common/craft-parts/reference/plugins/python_plugin.rst",
    "common/craft-parts/reference/plugins/qmake_plugin.rst",
    "common/craft-parts/reference/plugins/rust_plugin.rst",
    "common/craft-parts/reference/plugins/scons_plugin.rst",
    "common/craft-parts/reference/plugins/go_plugin.rst",
    "common/craft-parts/reference/plugins/go_use_plugin.rst",
    "common/craft-parts/reference/plugins/uv_plugin.rst",
    # Extra non-craft-parts exclusions can be added after this comment
    "reuse/reference/extensions/integrations.rst",
    "reuse/reference/extensions/environment_variables.rst",
    "reuse/reference/extensions/environment_variables_spring_boot.rst",
    "reuse/tutorial/*"
]

rst_epilog = """
.. include:: /reuse/links.txt
"""

# MyST-specific settings

suppress_warnings = ["myst.domains"]  # Suppresses false warnings about missing methods
myst_heading_anchors = 4  # Autogenerates anchors for MD headings

autodoc_default_options = {"exclude-members": "model_post_init"}

# region Options for extensions
# Intersphinx extension
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "craft-parts": ("https://canonical-craft-parts.readthedocs-hosted.com/en/latest/", None),
    "juju": ("https://documentation.ubuntu.com/juju/3.6/", None),
    "ops": ("https://documentation.ubuntu.com/ops/latest/", None),
    "rockcraft": ("https://documentation.ubuntu.com/rockcraft/stable/", None),
    "12-factor": ("https://canonical-12-factor-app-support.readthedocs-hosted.com/latest/", None),
}
# See also:
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#confval-intersphinx_disabled_reftypes
intersphinx_disabled_reftypes = ["*"]

# Type hints configuration
set_type_checking_flag = True
typehints_fully_qualified = False
always_document_param_types = True

# Github config
github_username = "canonical"
github_repository = "charmcraft"

html_domain_indices = True

# Client-side page redirects.
rediraffe_redirects = "redirects.txt"

# endregion

# Setup libraries documentation snippets for use in charmcraft docs.
# TODO: This needs a craft-parts update first.
# common_docs_path = pathlib.Path(__file__).parent / "common"
# craft_parts_docs_path = pathlib.Path(craft_parts_docs.__file__).parent / "craft-parts"
# (common_docs_path / "craft-parts").unlink(missing_ok=True)
# (common_docs_path / "craft-parts").symlink_to(
#     craft_parts_docs_path, target_is_directory=True
# )


def generate_cli_docs(nil):
    gen_cli_docs_path = (project_dir / "tools/gen_cli_docs.py").resolve()
    subprocess.run([sys.executable, gen_cli_docs_path, project_dir / "docs"], check=True)


def setup(app):
    app.connect("builder-inited", generate_cli_docs)


# Setup libraries documentation snippets for use in charmcraft docs.
common_docs_path = pathlib.Path(__file__).parent / "common"
craft_parts_docs_path = pathlib.Path(craft_parts_docs.__file__).parent / "craft-parts"
(common_docs_path / "craft-parts").unlink(missing_ok=True)
(common_docs_path / "craft-parts").symlink_to(craft_parts_docs_path, target_is_directory=True)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Charmcraft-specific poetry plugin."""

import pathlib
import shlex
from pathlib import Path

from craft_parts.plugins import poetry_plugin
from overrides import override

from charmcraft import utils


class PoetryPluginProperties(poetry_plugin.PoetryPluginProperties, frozen=True):
    poetry_keep_bins: bool = False
    """Keep the virtual environment's 'bin' directory."""


class PoetryPlugin(poetry_plugin.PoetryPlugin):
    """Charmcraft-specific version of the poetry plugin."""

    properties_class = PoetryPluginProperties
    _options: PoetryPluginProperties  # type: ignore[reportIncompatibleVariableOverride]

    def get_build_environment(self) -> dict[str, str]:
        return utils.extend_python_build_environment(super().get_build_environment())

    def _get_venv_directory(self) -> Path:
        return self._part_info.part_install_dir / "venv"

    def _get_pip(self) -> str:
        """Get the pip command to use."""
        return f"{self._get_system_python_interpreter()} -m pip --python=${{PARTS_PYTHON_VENV_INTERP_PATH}}"

    def _get_pip_install_commands(self, requirements_path: pathlib.Path) -> list[str]:
        """Get the commands for installing with pip.

        This only installs the dependencies from requirements, unlike the upstream
        version, because charms are not installable Python packages.

        :param requirements_path: The path of the requirements.txt file to write to.
        :returns: A list of strings forming the install script.
        """
        pip = self._get_pip()
        pip_extra_args = shlex.join(self._options.poetry_pip_extra_args)
        return [
            # These steps need to be separate because poetry export defaults to including
            # hashes, which don't work with installing from a directory.
            f"{pip} install --no-deps --no-binary=:all: {pip_extra_args} '--requirement={requirements_path}'",
            # Check that the virtualenv is consistent.
            f"{pip} check",
        ]

    def _get_package_install_commands(self) -> list[str]:
        """Get the package installation commands.

        This overrides the generic class to also:

        1. Copy the charm source into the charm.
        2. Copy the charmlibs into the charm.
        """
        return [
            *super()._get_package_install_commands(),
            *utils.get_charm_copy_commands(
                self._part_info.part_build_dir, self._part_info.part_install_dir
            ),
        ]

    def _should_remove_symlinks(self) -> bool:
        return True

    def _get_rewrite_shebangs_commands(self) -> list[str]:
        """Get the commands used to rewrite shebangs in the install dir.

        Charms don't need the shebangs to be rewritten.
        """
        return []

    @override
    def get_build_commands(self) -> list[str]:
        """Get the build commands for the Python plugin."""
        return [
            *super().get_build_commands(),
            *utils.get_venv_cleanup_commands(
                self._get_venv_directory(), keep_bins=self._options.poetry_keep_bins
            ),
        ]



# Adds database to Flask application

import os

import psycopg2


DATABASE_URI = os.environ["POSTGRESQL_DB_CONNECT_STRING"]


def migrate():
    with psycopg2.connect(DATABASE_URI) as conn, conn.cursor() as cur:
        cur.execute("""
            CREATE TABLE IF NOT EXISTS visitors (
                timestamp TIMESTAMP NOT NULL,
                user_agent TEXT NOT NULL
            );
        """)
        conn.commit()


if __name__ == "__main__":
    migrate()



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Charmcraft-specific poetry plugin."""

import shlex
from pathlib import Path

from craft_parts.plugins import python_plugin
from overrides import override

from charmcraft import utils


class PythonPluginProperties(python_plugin.PythonPluginProperties, frozen=True):
    python_packages: list[str] = []  # No default packages.
    python_keep_bins: bool = False
    """Keep the virtual environment's 'bin' directory."""


class PythonPlugin(python_plugin.PythonPlugin):
    """Charmcraft-specific version of the python plugin."""

    properties_class = PythonPluginProperties
    _options: PythonPluginProperties  # type: ignore[reportIncompatibleVariableOverride]

    @override
    def get_build_environment(self) -> dict[str, str]:
        return {
            "PIP_NO_BINARY": ":all:",
        } | utils.extend_python_build_environment(super().get_build_environment())

    @override
    def _get_venv_directory(self) -> Path:
        return self._part_info.part_install_dir / "venv"

    @override
    def _get_pip(self) -> str:
        """Get the pip command to use."""
        return f"{self._get_system_python_interpreter()} -m pip --python=${{PARTS_PYTHON_VENV_INTERP_PATH}}"

    @override
    def _get_package_install_commands(self) -> list[str]:
        """Get the package installation commands.

        This overrides the generic class in the following ways:

        1. Doesn't try to install '.' (charms are not installable packages)
        2. Copy the charm source into the charm.
        3. Copy the charmlibs into the charm.
        """
        pip = self._get_pip()
        install_params = shlex.join(
            (
                *(
                    f"--constraint={constraint}"
                    for constraint in self._options.python_constraints
                ),
                *(
                    f"--requirement={requirement}"
                    for requirement in self._options.python_requirements
                ),
                *self._options.python_packages,
            )
        )
        return [
            f"{pip} install --no-deps {install_params}",
            f"{pip} check",
            *utils.get_charm_copy_commands(
                self._part_info.part_build_dir, self._part_info.part_install_dir
            ),
        ]

    @override
    def _should_remove_symlinks(self) -> bool:
        return True

    @override
    def _get_rewrite_shebangs_commands(self) -> list[str]:
        """Get the commands used to rewrite shebangs in the install dir.

        Charms don't need the shebangs to be rewritten.
        """
        return []

    @override
    def get_build_commands(self) -> list[str]:
        """Get the build commands for the Python plugin."""
        return [
            *super().get_build_commands(),
            *utils.get_venv_cleanup_commands(
                self._get_venv_directory(), keep_bins=self._options.python_keep_bins
            ),
        ]



#!/usr/bin/env python3
"""Charm the application."""

import logging

import ops

# A standalone module for workload-specific logic (no charming concerns):
import test_cmd

logger = logging.getLogger(__name__)


class TestCmdCharm(ops.CharmBase):
    """Charm the application."""

    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        framework.observe(self.on.install, self._on_install)
        framework.observe(self.on.start, self._on_start)

    def _on_install(self, event: ops.InstallEvent):
        """Install the workload on the machine."""
        test_cmd.install()

    def _on_start(self, event: ops.StartEvent):
        """Handle start event."""
        self.unit.status = ops.MaintenanceStatus("starting workload")
        test_cmd.start()
        version = test_cmd.get_version()
        if version is not None:
            self.unit.set_workload_version(version)
        self.unit.status = ops.ActiveStatus()


if __name__ == "__main__":  # pragma: nocover
    ops.main(TestCmdCharm)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Module for helping with creating a dispatch script for charms."""

import pathlib

import craft_cli

from charmcraft import const

DISPATCH_SCRIPT_TEMPLATE = """\
#!/bin/sh
dispatch_path="$(dirname $(realpath $0))"
venv_bin_path="${{dispatch_path}}/venv/bin"
python_path="${{venv_bin_path}}/python"
if [ ! -e "${{python_path}}" ]; then
    mkdir -p "${{venv_bin_path}}"
    ln -s $(which python3) "${{python_path}}"
fi

# Add charm lib and source directories to PYTHONPATH so the charm can import
# libraries and its own modules as expected.
export PYTHONPATH="${{dispatch_path}}/lib:${{dispatch_path}}/src"

# Add the charm's lib and usr/lib directories to LD_LIBRARY_PATH, allowing
# staged packages to be discovered by the dynamic linker.
export LD_LIBRARY_PATH="${{dispatch_path}}/usr/lib:${{dispatch_path}}/lib:${{dispatch_path}}/usr/lib/$(uname -m)-linux-gnu"

exec "${{python_path}}" "${{dispatch_path}}/{entrypoint}"
"""


def create_dispatch(
    *, prime_dir: pathlib.Path, entrypoint: str = "src/charm.py"
) -> bool:
    """If the charm has no hooks or dispatch, create a dispatch file.

    :param prime_dir: the prime directory to inspect and create the file in.
    :returns: True if the file was created, False otherwise.
    """
    dispatch_path = prime_dir / const.DISPATCH_FILENAME
    hooks_path = prime_dir / const.HOOKS_DIRNAME

    if hooks_path.is_dir() or dispatch_path.is_file():
        return False

    if not (prime_dir / entrypoint).exists():
        return False

    craft_cli.emit.progress("Creating dispatch file")
    dispatch_path.write_text(DISPATCH_SCRIPT_TEMPLATE.format(entrypoint=entrypoint))
    dispatch_path.chmod(mode=0o755)

    return True



# FastAPI application that keeps track of visitors using a database

import datetime
import os
from typing import Annotated

from fastapi import FastAPI, Header
import psycopg2

app = FastAPI()
DATABASE_URI = os.environ["POSTGRESQL_DB_CONNECT_STRING"]


@app.get("/")
async def root(user_agent: Annotated[str | None, Header()] = None):
    with psycopg2.connect(DATABASE_URI) as conn, conn.cursor() as cur:
        timestamp = datetime.datetime.now()

        cur.execute(
            "INSERT INTO visitors (timestamp, user_agent) VALUES (%s, %s)",
            (timestamp, user_agent)
        )
        conn.commit()

    return {"message": os.getenv("APP_GREETING", "Hello World")}


@app.get("/visitors")
async def visitors():
    with psycopg2.connect(DATABASE_URI) as conn, conn.cursor() as cur:
        cur.execute("SELECT COUNT(*) FROM visitors")
        total_visitors = cur.fetchone()[0]

    return {"count": total_visitors}



# Copyright 2023,2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import sys
from typing import Any

import pytest
from overrides import override

from charmcraft import extensions
from tests.extensions.test_extensions import FakeExtension


class MyFakeExtension(FakeExtension):
    """A fake test Extension that has complete behavior"""

    name = "test-extension"
    bases = [("ubuntu", "22.04")]

    @override
    def get_root_snippet(self) -> dict[str, Any]:
        """Return the root snippet to apply."""
        return {"terms": ["https://example.com/test"]}


# Add an extension to the project YAML.
@pytest.fixture
def fake_project_yaml(fake_project_yaml):
    return fake_project_yaml + "\nextensions: [test-extension]"


@pytest.fixture
def fake_extensions(stub_extensions):
    extensions.register(MyFakeExtension.name, MyFakeExtension)


def test_expand_extensions_simple(
    fake_project_yaml, monkeypatch, new_path, app, fake_extensions, emitter
):
    """Expand a charmcraft.yaml with a single extension."""
    monkeypatch.setattr(sys, "argv", ["charmcraft", "expand-extensions"])
    project = app.services.get("project").get()

    app.run()

    assert project.terms == ["https://example.com/test"]

    emitter.assert_message(project.to_yaml_string())



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for store helpers."""

from hypothesis import given, strategies

from charmcraft import utils


@given(
    charms=strategies.lists(strategies.text()),
    bundles=strategies.lists(strategies.text()),
)
def test_get_packages(charms, bundles):
    packages = utils.get_packages(charms=charms, bundles=bundles)
    result_names = [package.package_name for package in packages]
    assert result_names == charms + bundles



# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Service classes charmcraft."""

from __future__ import annotations

from craft_application import ServiceFactory

# Add new services to this mapping to add them to the service factory
# Internal service name : Stringified service class name
_SERVICES: dict[str, str] = {
    "analysis": "AnalysisService",
    "build_plan": "CharmBuildPlanService",
    "charm_libs": "CharmLibsService",
    "image": "ImageService",
    "lifecycle": "LifecycleService",
    "package": "PackageService",
    "project": "ProjectService",
    "provider": "ProviderService",
    "remote_build": "RemoteBuildService",
    "store": "StoreService",
}


def register_services() -> None:
    """Register Charmcraft-specific services."""
    for name, service in _SERVICES.items():
        module_name = name.replace("_", "")
        ServiceFactory.register(
            name, service, module=f"charmcraft.services.{module_name}"
        )



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for store commands."""

import argparse
import datetime
import json
import pathlib
import re
from unittest import mock

import craft_cli.pytest_plugin
import pytest
from craft_store import publisher

from charmcraft import errors, utils
from charmcraft.application.commands import FetchLibCommand
from charmcraft.application.commands.store import CreateTrack
from tests import factory

OPERATOR_LIBS_LINUX_APT_ID = "7c3dbc9c2ad44a47bd6fcb25caa270e5"
OPERATOR_LIBS_LINUX_SNAP_ID = "05394e5893f94f2d90feb7cbe6b633cd"
MYSQL_MYSQL_ID = "8c1428f06b1b4ec8bf98b7d980a38a8c"


# region fetch-lib tests
@pytest.mark.slow
@pytest.mark.parametrize("formatted", [None, "json"])
def test_fetchlib_simple_downloaded(
    emitter: craft_cli.pytest_plugin.RecordingEmitter,
    new_path: pathlib.Path,
    config,
    formatted: str | None,
):
    """Happy path fetching the lib for the first time (downloading it)."""
    saved_file = new_path / utils.get_lib_path("operator_libs_linux", "apt", 0)
    args = argparse.Namespace(
        library="charms.operator_libs_linux.v0.apt", format=formatted
    )
    FetchLibCommand(config).run(args)

    assert saved_file.exists()

    message = emitter.interactions[-1].args[1]

    if formatted:
        message_dict = json.loads(message)[0]
        assert isinstance(message_dict["fetched"]["patch"], int)
        assert len(message_dict["fetched"]["content_hash"]) == 64  # sha256 hash
        del message_dict["fetched"]
        assert message_dict == {
            "charm_name": "operator-libs-linux",
            "library_name": "apt",
            "library_id": OPERATOR_LIBS_LINUX_APT_ID,
            "api": 0,
        }
    else:
        assert re.match(
            r"Library charms\.operator_libs_linux\.v0\.apt version 0.[0-9]+ downloaded.",
            message,
        )

    lib = utils.get_lib_info(lib_path=saved_file)
    assert lib.api == 0
    assert lib.charm_name == "operator-libs-linux"
    assert lib.lib_name == "apt"
    assert lib.lib_id == OPERATOR_LIBS_LINUX_APT_ID
    assert lib.patch > 1


@pytest.mark.slow
def test_fetchlib_simple_updated(
    emitter: craft_cli.pytest_plugin.RecordingEmitter, new_path: pathlib.Path, config
):
    """Happy path fetching the lib for Nth time (updating it)."""
    content, content_hash = factory.create_lib_filepath(
        "operator-libs-linux", "apt", api=0, patch=1, lib_id=OPERATOR_LIBS_LINUX_APT_ID
    )

    args = argparse.Namespace(library="charms.operator_libs_linux.v0.apt", format=None)
    FetchLibCommand(config).run(args)

    message = emitter.interactions[-1].args[1]

    assert re.match(
        r"Library charms\.operator_libs_linux\.v0\.apt updated to version 0\.[0-9]+\.",
        message,
    )

    saved_file = new_path / utils.get_lib_path("operator_libs_linux", "apt", 0)
    lib = utils.get_lib_info(lib_path=saved_file)
    assert lib.api == 0
    assert lib.charm_name == "operator-libs-linux"
    assert lib.lib_name == "apt"
    assert lib.lib_id == OPERATOR_LIBS_LINUX_APT_ID
    assert lib.patch > 1


@pytest.mark.slow
@pytest.mark.parametrize("formatted", [None, "json"])
def test_fetchlib_all(
    emitter: craft_cli.pytest_plugin.RecordingEmitter,
    new_path: pathlib.Path,
    config,
    formatted: str | None,
):
    """Update all the libraries found in disk."""
    factory.create_lib_filepath(
        "operator-libs-linux",
        "snap",
        api=0,
        patch=1,
        lib_id=OPERATOR_LIBS_LINUX_SNAP_ID,
    )
    factory.create_lib_filepath("mysql", "mysql", api=0, patch=1, lib_id=MYSQL_MYSQL_ID)

    args = argparse.Namespace(library=None, format=formatted)
    FetchLibCommand(config).run(args)
    message = emitter.interactions[-1].args[1]

    if formatted:
        message_list = json.loads(message)
        for message_dict in message_list:
            assert isinstance(message_dict["fetched"]["patch"], int)
            assert len(message_dict["fetched"]["content_hash"]) == 64  # sha256 hash
            del message_dict["fetched"]
        assert message_list == [
            {
                "charm_name": "mysql",
                "library_name": "mysql",
                "library_id": MYSQL_MYSQL_ID,
                "api": 0,
            },
            {
                "charm_name": "operator-libs-linux",
                "library_name": "snap",
                "library_id": OPERATOR_LIBS_LINUX_SNAP_ID,
                "api": 0,
            },
        ]
    else:
        assert re.match(
            r"Library charms\.[a-z_]+\.v0\.[a-z]+ updated to version 0\.[0-9]+\.",
            message,
        )

    saved_file = new_path / utils.get_lib_path("operator_libs_linux", "snap", 0)
    lib = utils.get_lib_info(lib_path=saved_file)
    assert lib.api == 0
    assert lib.charm_name == "operator-libs-linux"
    assert lib.lib_name == "snap"
    assert lib.lib_id == OPERATOR_LIBS_LINUX_SNAP_ID
    assert lib.patch > 1

    saved_file = new_path / utils.get_lib_path("mysql", "mysql", 0)
    lib = utils.get_lib_info(lib_path=saved_file)
    assert lib.api == 0
    assert lib.charm_name == "mysql"
    assert lib.lib_name == "mysql"
    assert lib.lib_id == MYSQL_MYSQL_ID
    assert lib.patch > 1


@pytest.mark.slow
@pytest.mark.parametrize("formatted", [None, "json"])
def test_fetchlib_store_not_found(
    emitter: craft_cli.pytest_plugin.RecordingEmitter,
    new_path: pathlib.Path,
    config,
    formatted: str | None,
) -> None:
    """The indicated library is not found in the store."""
    args = argparse.Namespace(library="charms.testcharm.v0.testlib", format=formatted)

    with pytest.raises(errors.LibraryError) as exc_info:
        FetchLibCommand(config).run(args)

    assert exc_info.value.args[0] == (
        "Library charms.testcharm.v0.testlib not found in Charmhub."
    )


@pytest.mark.slow
@pytest.mark.parametrize("formatted", [None, "json"])
def test_fetchlib_store_is_old(
    emitter: craft_cli.pytest_plugin.RecordingEmitter,
    new_path: pathlib.Path,
    config,
    formatted: str | None,
):
    """The store has an older version that what is found locally."""
    factory.create_lib_filepath(
        "mysql", "mysql", api=0, patch=2**63, lib_id=MYSQL_MYSQL_ID
    )

    args = argparse.Namespace(library="charms.mysql.v0.mysql", format=formatted)
    FetchLibCommand(config).run(args)

    error_message = (
        "Library charms.mysql.v0.mysql has local changes, cannot be updated."
    )
    if formatted:
        expected = [
            {
                "charm_name": "mysql",
                "library_name": "mysql",
                "library_id": MYSQL_MYSQL_ID,
                "api": 0,
                "error_message": error_message,
            },
        ]
        emitter.assert_json_output(  # pyright: ignore[reportAttributeAccessIssue]
            expected
        )
    else:
        emitter.assert_message(error_message)


@pytest.mark.slow
def test_fetchlib_store_same_versions_same_hash(
    emitter: craft_cli.pytest_plugin.RecordingEmitter, new_path: pathlib.Path, config
):
    """The store situation is the same than locally."""
    args = argparse.Namespace(library="charms.operator_libs_linux.v0.snap", format=None)
    # This run is a setup run
    FetchLibCommand(config).run(args)

    # The real run
    FetchLibCommand(config).run(args)

    assert re.match(
        r"Library charms.operator_libs_linux.v0.snap was already up to date in version 0.[0-9]+.",
        emitter.interactions[-1].args[1],
    )


@pytest.mark.slow
def test_fetchlib_store_same_versions_different_hash(
    emitter: craft_cli.pytest_plugin.RecordingEmitter, new_path: pathlib.Path, config
):
    """The store has the lib in the same version, but with different content."""
    args = argparse.Namespace(library="charms.operator_libs_linux.v0.snap", format=None)
    lib_path = utils.get_lib_path("operator-libs-linux", "snap", 0)
    # This run is a setup run
    FetchLibCommand(config).run(args)
    with lib_path.open("a+") as f:
        f.write("# This changes the hash!")

    # The real run
    FetchLibCommand(config).run(args)

    assert emitter.interactions[-1].args[1] == (
        "Library charms.operator_libs_linux.v0.snap has local changes, cannot be updated."
    )


# endregion


def test_create_track(emitter, service_factory, config):
    cmd = CreateTrack(config)
    args = argparse.Namespace(
        name="my-charm",
        track=["my-track"],
        automatic_phasing_percentage=None,
        format="json",
    )
    mock_create_tracks = mock.Mock()
    track = publisher.Track.unmarshal(
        {
            "name": "my-track",
            "automatic-phasing-percentage": None,
            "created-at": datetime.datetime.now(),
        }
    )
    mock_get_package_metadata = mock.Mock(
        return_value=publisher.RegisteredName.unmarshal(
            {
                "id": "mentalism",
                "private": False,
                "publisher": {"id": "EliBosnick"},
                "status": "hungry",
                "store": "charmhub",
                "type": "charm",
                "tracks": [
                    track,
                    publisher.Track.unmarshal(
                        {
                            "name": "latest",
                            "automatic-phasing-percentage": None,
                            "created-at": datetime.datetime.now(),
                        }
                    ),
                ],
            }
        )
    )

    service_factory.store._publisher.create_tracks = mock_create_tracks
    service_factory.store._publisher.get_package_metadata = mock_get_package_metadata

    cmd.run(args)

    emitter.assert_json_output(
        [{"name": "my-track", "automatic-phasing-percentage": None}]
    )



#!/usr/bin/env python3
# Copyright 2025 Ubuntu
# See LICENSE file for licensing details.

"""Flask Charm entrypoint."""

import logging
import typing

import ops

import paas_charm.flask

import requests

logger = logging.getLogger(__name__)


class FlaskHelloWorldCharm(paas_charm.flask.Charm):
    """Flask Charm service."""

    def __init__(self, *args: typing.Any) -> None:
        """Initialize the instance.

        Args:
            args: passthrough to CharmBase.
        """
        super().__init__(*args)
        self.framework.observe(self.on.updatelogfile_action,
                               self._on_updatelogfile_action)

    def _on_updatelogfile_action(self, event: ops.ActionEvent) -> None:
        """Handle the updatelogfile action.

        Args:
            event: the action event object.
            logfile: the output logfile in the container
        """
        if not self.is_ready():
            event.fail("flask-app container is not ready")
        try:
            response = requests.get(
                    f"http://127.0.0.1:{self._workload_config.port}", timeout=5
                    )
            response.raise_for_status()
            # push response to file in app container
            self._container.push(event.params["logfile"], response.text)
            output = "App response: " + response.text
            # access file in container and read its contents
            # (you could put this part in a separate action called readlogfile)
            output_comp = self._container.pull(event.params["logfile"]).read()
            output += "Output written to file: " + output_comp
            event.set_results({"result": output})
        except ops.pebble.PathError as e:
            event.fail(str(e.message))
        except requests.exceptions.RequestException as e:
            # if it failed with http bad status code or the connection failed
            if e.response is None:
                event.fail(
                    f"unable to connect on port {self._workload_config.port}"
                    )
            else:
                event.fail(
                    f"workload responded with code {e.response.status_code}"
                    )

if __name__ == "__main__":
    ops.main(FlaskHelloWorldCharm)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Service class for creating providers."""

from __future__ import annotations

import contextlib
import fcntl
import io
import os
import pathlib
from collections.abc import Callable, Generator
from typing import cast

import craft_application
import craft_platforms
import craft_providers
from craft_application import services
from craft_cli import emit
from craft_providers import bases
from typing_extensions import override

from charmcraft import env


class ProviderService(services.ProviderService):
    """Business logic for getting providers."""

    def __init__(
        self,
        app: craft_application.AppMetadata,
        services: craft_application.ServiceFactory,
        *,
        work_dir: pathlib.Path,
        provider_name: str | None = None,
        install_snap: bool = True,
    ) -> None:
        super().__init__(
            app,
            services,
            work_dir=work_dir,
            provider_name=provider_name,
            install_snap=install_snap,
        )
        self._cache_path: pathlib.Path | None = None
        self._lock: io.TextIOBase | None = None

    def setup(self) -> None:
        """Set up the provider service for Charmcraft."""
        super().setup()

        # Forward all charmcraft environment variables
        for key, value in os.environ.items():
            if key.startswith("CHARMCRAFT_"):
                self.environment[key] = value

        self.environment["CHARMCRAFT_MANAGED_MODE"] = "1"

    def get_base(
        self,
        base_name: bases.BaseName,
        *,
        instance_name: str,
        **kwargs: bool | str | None | pathlib.Path,
    ) -> craft_providers.Base:
        """Get the base configuration from a base name.

        :param base_name: The base to lookup.
        :param instance_name: A name to assign to the instance.
        :param kwargs: Additional keyword arguments are sent directly to the base.

        If no cache_path is included, adds one.
        """
        self._cache_path = cast(
            pathlib.Path, kwargs.get("cache_path", env.get_host_shared_cache_path())
        )
        self._lock = _maybe_lock_cache(self._cache_path)

        # Forward the shared cache path.
        kwargs["cache_path"] = self._cache_path if self._lock else None
        return super().get_base(
            base_name,
            instance_name=instance_name,
            # craft-application annotation is incorrect
            **kwargs,  # type: ignore[arg-type]
        )

    @override
    @contextlib.contextmanager
    def instance(
        self,
        build_info: craft_platforms.BuildInfo,
        *,
        work_dir: pathlib.Path,
        allow_unstable: bool = True,
        clean_existing: bool = False,
        use_base_instance: bool = True,
        project_name: str | None = None,
        prepare_instance: Callable[[craft_providers.Executor], None] | None = None,
        **kwargs: bool | str | None,
    ) -> Generator[craft_providers.Executor, None, None]:
        """Instance override for Charmcraft."""
        with super().instance(
            build_info,
            work_dir=work_dir,
            allow_unstable=allow_unstable,
            clean_existing=clean_existing,
            prepare_instance=prepare_instance,
            project_name=project_name,
            **kwargs,  # type: ignore[arg-type]
        ) as instance:
            try:
                instance.execute_run(["chmod", "a+rwx", "/tmp/craft-state"])
                # Use /root/.cache even if we're in the snap.
                instance.execute_run(
                    ["rm", "-rf", "/root/snap/charmcraft/common/cache"], check=True
                )
                instance.execute_run(["mkdir", "-p", "/root/.cache"], check=True)
                instance.execute_run(
                    ["ln", "-s", "/root/.cache", "/root/snap/charmcraft/common/cache"],
                    check=True,
                )
                yield instance
            finally:
                if self._lock:
                    fcntl.flock(self._lock, fcntl.LOCK_UN)
                    self._lock.close()


def _maybe_lock_cache(path: pathlib.Path) -> io.TextIOBase | None:
    """Lock the cache so we only have one copy of Charmcraft using it at a time."""
    cache_lock_path = path / "charmcraft.lock"

    emit.trace("Attempting to lock the cache path")
    lock_file = cache_lock_path.open("w+")
    try:
        # Exclusive lock, but non-blocking.
        fcntl.flock(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except OSError:
        emit.progress(
            "Shared cache locked by another process; running without cache.",
            permanent=True,
        )
        return None
    else:
        pid = str(os.getpid())
        lock_file.write(pid)
        lock_file.flush()
        os.fsync(lock_file.fileno())
        emit.trace(f"Cache path locked by this process ({pid})")
        return lock_file



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""craft-application based lifecycle commands."""

from __future__ import annotations

import pathlib
import textwrap
from typing import TYPE_CHECKING, Any, cast

import craft_cli
from craft_application.commands import lifecycle
from craft_application.util import is_managed_mode
from craft_cli import CraftError
from typing_extensions import override

from charmcraft import models, utils

if TYPE_CHECKING:  # pragma: no cover
    import argparse

    from charmcraft.services.charmlibs import CharmLibsService
    from charmcraft.services.store import StoreService


def get_lifecycle_commands() -> list[type[craft_cli.BaseCommand]]:
    """Return the lifecycle related command group."""
    return [
        lifecycle.CleanCommand,
        lifecycle.PullCommand,
        lifecycle.BuildCommand,
        lifecycle.StageCommand,
        lifecycle.PrimeCommand,
        PackCommand,
        lifecycle.TestCommand,
    ]


class PackCommand(lifecycle.PackCommand):
    """Command to pack the final artifact."""

    name = "pack"
    help_msg = "Build the charm"
    overview = textwrap.dedent(
        """
        Build and pack a charm operator.

        You can `juju deploy` the resulting `.charm` file directly, or
        upload it to Charmhub with `charmcraft upload`.

        For the charm you must be inside a charm directory with a valid
        `metadata.yaml`, `requirements.txt` including the `ops` package
        for the Python operator framework, and an operator entrypoint,
        usually `src/charm.py`.  See `charmcraft init` to create a
        template charm directory structure.
        """
    )

    @override
    def _fill_parser(self, parser: argparse.ArgumentParser) -> None:
        super()._fill_parser(parser)

        parser.add_argument(
            "--bases-index",
            action="append",
            type=int,
            help="Index of 'bases' configuration to build (can be used multiple "
            "times); zero-based, defaults to all",
        )
        parser.add_argument(
            "--force",
            action="store_true",
            help="Force packing even after finding lint errors",
        )
        parser.add_argument(
            "--measure",
            type=pathlib.Path,
            help="Dump measurements to the specified file",
        )
        parser.add_argument(
            "--format",
            choices=["json"],
            help="Produce a machine-readable format (currently only json)",
        )
        parser.add_argument(
            "--project-dir",
            "-p",
            type=pathlib.Path,
            default=pathlib.Path.cwd(),
            help="Specify the project's directory (defaults to current)",
        )

    @override
    @staticmethod
    def _should_add_shell_args() -> bool:
        return True

    def _validate_bases_indices(self, bases_indices):
        """Validate that bases index is valid."""
        if bases_indices is None:
            return

        project = cast(models.Charm, self._services.project)

        msg = "Bases index '{}' is invalid (must be >= 0 and fit in configured bases)."
        len_configured_bases = len(project.bases)
        for bases_index in bases_indices:
            if bases_index < 0:
                raise CraftError(msg.format(bases_index))
            if bases_index >= len_configured_bases:
                raise CraftError(msg.format(bases_index))

    def _update_charm_libs(self) -> None:
        """Update charm libs attached to the project."""
        craft_cli.emit.progress(
            "Checking that charmlibs match 'charmcraft.yaml' values"
        )
        project = cast(models.CharmcraftProject, self._services.get("project").get())
        libs_svc = cast("CharmLibsService", self._services.get("charm_libs"))
        installable_libs: list[models.CharmLib] = []
        for lib in project.charm_libs:
            library_name = utils.QualifiedLibraryName.from_string(lib.lib)
            if not libs_svc.get_local_version(
                charm_name=library_name.charm_name, lib_name=library_name.lib_name
            ):
                installable_libs.append(lib)
        if installable_libs:
            store = cast("StoreService", self._services.store)
            libraries_md = store.get_libraries_metadata(installable_libs)
            with craft_cli.emit.progress_bar(
                "Downloading charmlibs...", len(installable_libs)
            ) as progress:
                for library in libraries_md:
                    craft_cli.emit.debug(repr(library))
                    lib_contents = store.get_library(
                        library.charm_name,
                        library_id=library.lib_id,
                        api=library.api,
                        patch=library.patch,
                    )
                    libs_svc.write(lib_contents)
                    progress.advance(1)

    def _run(
        self,
        parsed_args: argparse.Namespace,
        step_name: str | None = None,
        **kwargs: Any,
    ) -> None:
        project = cast(models.CharmcraftProject, self._services.get("project").get())
        if project.charm_libs:
            self._update_charm_libs()

        result = super()._run(parsed_args, step_name, **kwargs)

        # Move artifacts in the outer instance.
        if not is_managed_mode():
            state_service = self._services.get("state")
            try:
                artifacts = cast(dict[str, pathlib.Path], state_service.get("artifact"))
            except KeyError:
                craft_cli.emit.debug(
                    "Could not find artifacts in the state service. Not moving."
                )
            else:
                project_dir = parsed_args.project_dir or pathlib.Path.cwd()
                output_dir = parsed_args.output or pathlib.Path.cwd()

                for artifact in artifacts.values():
                    old_path = project_dir / artifact
                    new_path = output_dir / artifact
                    if old_path != new_path:
                        old_path.rename(new_path)

        return result



# Copyright 2020-2022,2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import itertools
import re
from pathlib import Path

import pytest

# A list of bash file globs to not check for
IGNORE_GLOBS = frozenset({"tests/spread/**/*.py"})


def get_python_filepaths() -> list[str]:
    """Helper to retrieve paths of Python files."""
    # list of directories to scan
    source_dirs = ["charmcraft", "tests"]
    # list of source files
    source_files = []

    # Parse the globs into their matching files
    cwd = Path.cwd()
    ignore_files: list[Path] = []
    for glob in IGNORE_GLOBS:
        ignore_files.extend(cwd.glob(glob))

    for source_dir in source_dirs:
        # Loop over the source_dir recursively
        # This is done instead of os.walk() to take advantage of Path.resolve()
        for file in Path(source_dir).resolve().glob("**/*.py"):
            if file in ignore_files:
                continue

            source_files.append(str(file))

    return source_files


def test_ensure_copyright() -> None:
    """Check that all non-empty Python files have copyright somewhere in the first 5 lines."""
    issues = []
    regex = re.compile(r"# Copyright \d{4}([-,]\d{4})* Canonical Ltd.$")
    for filepath in get_python_filepaths():
        if Path(filepath).stat().st_size == 0:
            continue
        if filepath.endswith("charmcraft/_version.py") or filepath.endswith(
            "charmcraft\\_version.py"
        ):
            continue

        with open(filepath, encoding="utf8") as fh:
            for line in itertools.islice(fh, 5):
                if regex.match(line):
                    break
            else:
                issues.append(filepath)
    if issues:
        msg = "Please add copyright headers to the following files:\n" + "\n".join(
            issues
        )
        pytest.fail(msg, pytrace=False)



# Copyright 2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Project service for Charmcraft."""

import itertools
import pathlib

import craft_platforms
from craft_application.services.project import ProjectService as BaseProjectService
from craft_cli import CraftError
from typing_extensions import Any, override

from charmcraft import const, extensions, preprocess
from charmcraft.models.charmcraft import BasesConfiguration


class ProjectService(BaseProjectService):
    """Charmcraft-specific project service."""

    @override
    def _app_render_legacy_platforms(self) -> dict[str, craft_platforms.PlatformDict]:
        """Convert a bases charm to a platforms charm."""
        raw = self.get_raw()
        if "bases" not in raw:
            return super()._app_render_legacy_platforms()

        bases_config = [
            BasesConfiguration.model_validate(base) for base in raw["bases"]
        ]

        platforms: dict[str, craft_platforms.PlatformDict] = {}
        invalid_build_bases: set[str] = set()

        for base in bases_config:
            platform_parts = [
                f"{run.name}-{run.channel}-{'-'.join(run.architectures)}"
                for run in base.run_on
            ]
            platform = "_".join(platform_parts)
            build_ons = list(
                itertools.chain.from_iterable(
                    build_on.to_strings() for build_on in base.build_on
                )
            )
            build_fors = list(
                itertools.chain.from_iterable(
                    build_for.to_strings() for build_for in base.run_on
                )
            )
            for build_on in base.build_on:
                if (build_str := build_on.to_os_string()) not in const.LEGACY_BASES:
                    invalid_build_bases.add(build_str)

            platforms[platform] = {
                "build-on": build_ons,
                "build-for": build_fors,
            }

        if invalid_build_bases:
            bases_str = ", ".join(f"'{base}'" for base in sorted(invalid_build_bases))
            brief = f"Not valid for use with the 'bases' key: {bases_str}"
            raise CraftError(
                brief,
                resolution="Use the 'platforms' keyword in order to use newer bases.",
            )

        return platforms

    @override
    @staticmethod
    def _app_preprocess_project(
        project: dict[str, Any],
        *,
        build_on: str,
        build_for: str,
        platform: str,
    ) -> None:
        """Run Charmcraft-specific pre-processing on the project."""
        # Extensions get applied on as close as possible to what the user provided.
        project_dir = pathlib.Path.cwd()
        extensions.apply_extensions(project_dir, project)
        # Preprocessing "magic" to create a fully-formed charm.
        preprocess.add_default_parts(project)
        preprocess.add_config(project_dir, project)
        preprocess.add_actions(project_dir, project)
        preprocess.add_metadata(project_dir, project)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for lifecycle commands."""

import argparse
import pathlib
from typing import TYPE_CHECKING, cast
from unittest import mock

import pytest
from craft_cli.pytest_plugin import RecordingEmitter

from charmcraft import application, models, services, utils
from charmcraft.application.commands import lifecycle
from charmcraft.store.models import Library

if TYPE_CHECKING:
    from charmcraft.models.project import CharmcraftProject
    from charmcraft.services.charmlibs import CharmLibsService


def get_namespace(
    *,
    bases_index=None,
    debug=False,
    destructive_mode=False,
    force=None,
    shell=False,
    shell_after=False,
    format=None,
    measure=None,
    include_all_charms: bool = False,
    include_charm: list[pathlib.Path] | None = None,
    output_bundle: pathlib.Path | None = None,
    parts: list[str] | None = None,
) -> argparse.Namespace:
    if bases_index is None:
        bases_index = []

    return argparse.Namespace(
        bases_index=bases_index,
        debug=debug,
        destructive_mode=destructive_mode,
        force=force,
        shell=shell,
        shell_after=shell_after,
        format=format,
        measure=measure,
        include_all_charms=include_all_charms,
        include_charm=include_charm,
        output_bundle=output_bundle,
        parts=parts,
    )


@pytest.fixture
def pack(service_factory: services.ServiceFactory) -> lifecycle.PackCommand:
    return lifecycle.PackCommand(
        {"app": application.APP_METADATA, "services": service_factory}
    )


def test_pack_update_charm_libs_empty(
    project_path: pathlib.Path,
    pack: lifecycle.PackCommand,
    emitter: RecordingEmitter,
    service_factory: services.ServiceFactory,
    mock_store_anonymous_client: mock.Mock,
    check,
):
    project = cast("CharmcraftProject", service_factory.get("project").get())
    project.charm_libs = [models.CharmLib(lib="my_charm.my_lib", version="0.1")]
    store_lib = Library("lib_id", "my_lib", "my_charm", 0, 1, "Lib contents", "hash")
    mock_store_anonymous_client.fetch_libraries_metadata.return_value = [store_lib]
    mock_store_anonymous_client.get_library.return_value = store_lib

    libs_service = cast("CharmLibsService", service_factory.get("charm_libs"))
    libs_service.write = mock.Mock(wraps=libs_service.write)

    pack._update_charm_libs()

    libs_service.write.assert_called_once_with(store_lib)

    with check():
        emitter.assert_debug(repr(store_lib))

    path = project_path / utils.get_lib_path("my_charm", "my_lib", 0)
    assert path.read_text() == "Lib contents"


def test_pack_update_charm_libs_no_update(
    fake_project_dir: pathlib.Path,
    pack: lifecycle.PackCommand,
    simple_charm,
    emitter: RecordingEmitter,
    service_factory: services.ServiceFactory,
    mock_store_anonymous_client: mock.Mock,
):
    simple_charm.charm_libs = [models.CharmLib(lib="my_charm.my_lib", version="0.1")]
    store_lib = Library("lib_id", "my_lib", "my_charm", 0, 1, "Lib contents", "hash")
    path = fake_project_dir / utils.get_lib_path("my_charm", "my_lib", 0)
    path.parent.mkdir(parents=True)
    path.write_text("LIBID='id'\nLIBAPI=0\nLIBPATCH=1")
    mock_store_anonymous_client.fetch_libraries_metadata.return_value = [store_lib]
    mock_store_anonymous_client.get_library.return_value = store_lib

    pack._update_charm_libs()

    with pytest.raises(AssertionError):
        emitter.assert_debug(repr(store_lib))

    assert path.read_text() != "Lib contents"


def test_pack_update_charm_libs_needs_update(
    fake_project_dir: pathlib.Path,
    pack: lifecycle.PackCommand,
    simple_charm,
    emitter: RecordingEmitter,
    service_factory: services.ServiceFactory,
    mock_store_anonymous_client: mock.Mock,
):
    simple_charm.charm_libs = [models.CharmLib(lib="my_charm.my_lib", version="0.2")]
    store_lib = Library("lib_id", "my_lib", "my_charm", 0, 2, "Lib contents", "hash")
    path = fake_project_dir / utils.get_lib_path("my_charm", "my_lib", 0)
    path.parent.mkdir(parents=True)
    path.write_text("LIBID='id'\nLIBAPI=0\nLIBPATCH=1")
    mock_store_anonymous_client.fetch_libraries_metadata.return_value = [store_lib]
    mock_store_anonymous_client.get_library.return_value = store_lib

    pack._update_charm_libs()

    with pytest.raises(AssertionError):
        emitter.assert_debug(repr(store_lib))

    assert path.read_text() != "Lib contents"



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

from pathlib import Path

import pytest
import pytest_check

from charmcraft.parts import plugins


def test_get_build_environment(uv_plugin: plugins.UvPlugin):
    env = uv_plugin.get_build_environment()

    assert env["PARTS_PYTHON_VENV_ARGS"] == "--without-pip"


def test_get_venv_directory(uv_plugin: plugins.UvPlugin, install_path: Path):
    assert uv_plugin._get_venv_directory() == install_path / "venv"


def test_get_package_install_commands(
    uv_plugin: plugins.UvPlugin, build_path: Path, install_path: Path
):
    copy_src_cmd = (
        f"cp --archive --recursive --reflink=auto {build_path}/src {install_path}"
    )

    copy_lib_cmd = (
        f"cp --archive --recursive --reflink=auto {build_path}/lib {install_path}"
    )

    default_commands = uv_plugin._get_package_install_commands()

    pytest_check.is_not_in(copy_src_cmd, default_commands)
    pytest_check.is_not_in(copy_lib_cmd, default_commands)

    (build_path / "src").mkdir(parents=True)

    pytest_check.equal(
        uv_plugin._get_package_install_commands(), [*default_commands, copy_src_cmd]
    )

    (build_path / "lib" / "charm").mkdir(parents=True)

    pytest_check.equal(
        uv_plugin._get_package_install_commands(),
        [*default_commands, copy_src_cmd, copy_lib_cmd],
    )

    (build_path / "src").rmdir()

    pytest_check.equal(
        uv_plugin._get_package_install_commands(), [*default_commands, copy_lib_cmd]
    )


def test_do_not_install_project(uv_plugin: plugins.UvPlugin) -> None:
    for command in uv_plugin._get_package_install_commands():
        if command.startswith("uv sync") and "--no-install-project" in command:
            break
    else:
        pytest.fail(reason="Charms should not be installed as projects.")



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import pathlib
import platform
import subprocess
import sys

import craft_application
import pytest

pytestmark = [
    pytest.mark.skipif(sys.platform != "linux", reason="craft-parts is linux-only")
]


@pytest.fixture(autouse=True)
def add_part(
    service_factory: craft_application.ServiceFactory, project_path: pathlib.Path
):
    service_factory.get("project").get().parts = {
        "my-charm": {
            "plugin": "uv",
            "source": str(project_path),
            "source-type": "local",
        }
    }


@pytest.fixture
def uv_project(project_path: pathlib.Path, monkeypatch) -> None:
    subprocess.run(
        [
            "uv",
            "init",
            "--name=test-charm",
            f"--python={platform.python_version()}",
            "--no-progress",
            "--no-workspace",
        ],
        cwd=project_path,
        check=True,
    )
    subprocess.run(["uv", "add", "ops"], cwd=project_path, check=True)
    monkeypatch.delenv("UV_FROZEN", raising=False)
    subprocess.run(
        [
            "uv",
            "lock",
        ],
        cwd=project_path,
        check=True,
    )
    source_dir = project_path / "src"
    source_dir.mkdir()
    (source_dir / "charm.py").write_text("# Charm file")


@pytest.mark.slow
@pytest.mark.usefixtures("uv_project")
def test_uv_plugin(
    service_factory: craft_application.ServiceFactory, tmp_path: pathlib.Path
):
    install_path = tmp_path / "parts" / "my-charm" / "install"
    stage_path = tmp_path / "stage"

    service_factory.lifecycle.run("stage")

    # Check that the part install directory looks correct.
    assert (install_path / "src" / "charm.py").read_text() == "# Charm file"
    assert (install_path / "venv" / "lib").is_dir()

    # Check that the stage directory looks correct.
    assert (stage_path / "src" / "charm.py").read_text() == "# Charm file"
    assert (stage_path / "venv" / "lib").is_dir()
    assert not (stage_path / "venv" / "lib64").is_symlink()



# Copyright 2020-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Infrastructure for the 'init' command."""

import argparse
import os
import pathlib
import re
from datetime import date

from craft_cli import CraftError, emit

from charmcraft.application.commands import base
from charmcraft.utils import get_templates_environment, make_executable

try:
    import pwd
except ImportError:
    pwd = None  # type: ignore[assignment]

# the available profiles and in which directory the template can be found
PROFILES = {
    "kubernetes": "init-kubernetes",
    "machine": "init-machine",
    "flask-framework": "init-flask-framework",
    "django-framework": "init-django-framework",
    "go-framework": "init-go-framework",
    "fastapi-framework": "init-fastapi-framework",
    "expressjs-framework": "init-expressjs-framework",
    "spring-boot-framework": "init-spring-boot-framework",
    "test-kubernetes": "test-kubernetes",
    "test-machine": "test-machine",
}
DEFAULT_PROFILE = "kubernetes"


_overview = """
Initialize a charm operator package tree and files.

This command will modify the directory to create the necessary files for a
charm operator package. By default it will work in the current directory.

Available profiles are:
    kubernetes:
        A basic Kubernetes charm with example container.

    machine:
        A basic charm but meant to be deployed in machine-based environments,
        without container requirements.

    django-framework:
        A basic Kubernetes charm for a 12-factor Django app.

    fastapi-framework:
        A basic Kubernetes charm for a 12-factor FastAPI app.

    flask-framework:
        A basic Kubernetes charm for a 12-factor Flask app.

    go-framework:
        A basic Kubernetes charm for a 12-factor Go app.

    spring-boot-framework:
        A basic Kubernetes charm for a 12-factor Spring Boot app.

Depending on the profile choice, Charmcraft will setup the following tree of
files and directories::

    .
    ├── charmcraft.yaml            - Charm build configuration
    ├── CONTRIBUTING.md            - Instructions for how to build and develop
    │                                your charm
    ├── LICENSE                    - Your charm license, we recommend Apache 2
    ├── pyproject.toml             - Configuration for testing, formatting and
    │                                linting tools. Specifies Python dependencies for
    │                                your charm if profile is 'kubernetes' or 'machine'
    ├── README.md                  - Frontpage for your charmhub.io/charm/
    ├── requirements.txt           - Python dependencies for your charm, with Ops,
    │                                created for 12-factor app profiles only
    ├── src
    │   ├── charm.py               - Python code that operates your charm's workload
    │   └── <workload>.py          - Standalone module for workload-specific logic,
    │                                created if profile is 'kubernetes' or 'machine'
    ├── tests
    │   ├── integration
    │   │   └── test_charm.py      - Integration tests
    │   └── unit
    │       └── test_charm.py      - Unit tests
    ├── tox.ini                    - Configuration for tox, the tool to run all tests
    ├── uv.lock                    - Specifies exact versions of Python dependencies,
    │                                created if profile is 'kubernetes' or 'machine'

You will need to edit at least charmcraft.yaml and README.md.

Your minimal operator code is in src/charm.py, which uses the 'ops' Python framework.
See https://documentation.ubuntu.com/ops/latest/. There are also some sample unit and
integration tests, which you can run using 'tox -e unit' and 'tox -e integration'.
"""


def _make_success_message(src_files: list[str]) -> str:
    src_files_str = "\n".join(src_files)
    return f"""\
Charmed operator package file and directory tree initialised.

Now edit the following package files to provide fundamental charm metadata
and other information:

charmcraft.yaml
{src_files_str}
README.md
"""


def _make_workload_module_name(charm_name: str) -> str:
    module_name = charm_name.replace("-", "_")
    generic_names = [  # put names with more components at the beginning of the list
        "k8s_charm",
        "k8s_operator",
        "machine_charm",
        "machine_operator",
        "vm_charm",
        "vm_operator",
        "charm",
        "operator",
        "k8s",
        "machine",
        "vm",
    ]
    if module_name in generic_names:
        return "workload"
    for generic_name in generic_names:
        generic_suffix = f"_{generic_name}"
        if module_name.endswith(generic_suffix):
            return module_name[: -len(generic_suffix)]
    return module_name


def _get_users_full_name_gecos() -> str | None:
    """Get user's full name from Gecos (/etc/passwd)."""
    try:
        return pwd.getpwuid(os.getuid()).pw_gecos.split(",", 1)[0]
    except KeyError:
        return None


class InitCommand(base.CharmcraftCommand):
    """Initialize a directory to be a charm project."""

    name = "init"
    help_msg = "Initialize a charm operator package tree and files"
    overview = _overview
    common = True

    def fill_parser(self, parser):
        """Specify command's specific parameters."""
        parser.add_argument(
            "--name", help="The name of the charm; defaults to the directory name"
        )
        parser.add_argument(
            "--author",
            help="The charm author; defaults to the current user name per GECOS",
        )
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="Initialize even if the directory is not empty (will not overwrite files)",
        )
        parser.add_argument(
            "--profile",
            choices=list(PROFILES),
            default=DEFAULT_PROFILE,
            help=f"Use the specified project profile (defaults to '{DEFAULT_PROFILE}')",
        )
        parser.add_argument(
            "-p",
            "--project-dir",
            type=pathlib.Path,
            default=pathlib.Path.cwd(),
            help="Specify the project's directory (defaults to current)",
        )

    def run(self, parsed_args: argparse.Namespace):
        """Execute command's actual functionality."""
        init_dirpath = parsed_args.project_dir.resolve()
        if not init_dirpath.exists():
            init_dirpath.mkdir(parents=True)
        elif any(init_dirpath.iterdir()) and not parsed_args.force:
            tpl = "{!r} is not empty (consider using --force to work on nonempty directories)"
            raise CraftError(tpl.format(str(init_dirpath)))
        emit.debug(f"Using project directory {str(init_dirpath)!r}")

        if parsed_args.author is None and pwd is not None:
            parsed_args.author = _get_users_full_name_gecos()

        if not parsed_args.author:
            raise CraftError(
                "Unable to automatically determine author's name, specify it with --author"
            )

        if not parsed_args.name:
            parsed_args.name = init_dirpath.name
            emit.debug(f"Set project name to '{parsed_args.name}'")

        if not re.match(r"[a-z][a-z0-9-]*[a-z0-9]$", parsed_args.name):
            raise CraftError(
                f"{parsed_args.name} is not a valid charm name. "
                "The name must start with a lowercase letter "
                "and contain only alphanumeric characters and hyphens."
            )

        context = {
            "name": parsed_args.name,
            "author": parsed_args.author,
            "year": date.today().year,
            "class_name": "".join(re.split(r"\W+", parsed_args.name.title())) + "Charm",
            "workload_module": _make_workload_module_name(parsed_args.name),
        }

        template_directory = PROFILES[parsed_args.profile]
        env = get_templates_environment(template_directory)

        executables = [
            "run_tests",
            "src/charm.py",
            "tests/spread/lib/tools/retry",
            "spread/.extension",
        ]
        src_files = ["src/charm.py"]
        for template_name in env.list_templates():
            if not template_name.endswith(".j2"):
                continue
            template = env.get_template(template_name)
            template_name = template_name[:-3]
            emit.debug(f"Rendering {template_name}")
            path = init_dirpath / template_name
            if path.exists():
                continue
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("wt", encoding="utf8") as fh:
                out = template.render(context)
                fh.write(out)
                if template_name in executables and os.name == "posix":
                    make_executable(fh)
                    emit.debug("  made executable")
            if path.name == "workload.py" and path.parent.name == "src":
                workload_module = context["workload_module"]
                workload_module_path = path.with_name(f"{workload_module}.py")
                path.rename(workload_module_path)
                src_files.append(f"src/{workload_module}.py")
        for line in _make_success_message(src_files).split("\n"):
            emit.message(line)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for the Charmcraft-specific poetry plugin."""

import pathlib

import pytest_check

from charmcraft.parts import plugins


def test_get_build_environment(
    poetry_plugin: plugins.PoetryPlugin, install_path: pathlib.Path
):
    env = poetry_plugin.get_build_environment()

    assert env["PARTS_PYTHON_VENV_ARGS"] == "--without-pip"


def test_get_venv_directory(
    poetry_plugin: plugins.PoetryPlugin, install_path: pathlib.Path
):
    assert poetry_plugin._get_venv_directory() == install_path / "venv"


def test_get_pip_install_commands(poetry_plugin: plugins.PoetryPlugin):
    poetry_plugin._get_pip = lambda: "/python -m pip"

    assert poetry_plugin._get_pip_install_commands(
        pathlib.Path("/my dir/reqs.txt")
    ) == [
        "/python -m pip install --no-deps --no-binary=:all:  '--requirement=/my dir/reqs.txt'",
        "/python -m pip check",
    ]


def test_get_package_install_commands(
    poetry_plugin: plugins.PoetryPlugin,
    build_path: pathlib.Path,
    install_path: pathlib.Path,
):
    copy_src_cmd = (
        f"cp --archive --recursive --reflink=auto {build_path}/src {install_path}"
    )
    copy_lib_cmd = (
        f"cp --archive --recursive --reflink=auto {build_path}/lib {install_path}"
    )

    # Check if no src or libs exist
    default_commands = poetry_plugin._get_package_install_commands()

    pytest_check.is_not_in(copy_src_cmd, default_commands)
    pytest_check.is_not_in(copy_lib_cmd, default_commands)

    # With a src directory
    (build_path / "src").mkdir(parents=True)

    pytest_check.equal(
        poetry_plugin._get_package_install_commands(), [*default_commands, copy_src_cmd]
    )

    # With both src and lib
    (build_path / "lib" / "charm").mkdir(parents=True)

    pytest_check.equal(
        poetry_plugin._get_package_install_commands(),
        [*default_commands, copy_src_cmd, copy_lib_cmd],
    )

    # With only lib
    (build_path / "src").rmdir()

    pytest_check.equal(
        poetry_plugin._get_package_install_commands(), [*default_commands, copy_lib_cmd]
    )


def test_get_rm_command(
    poetry_plugin: plugins.PoetryPlugin, install_path: pathlib.Path
):
    assert (
        f"rm -rf {install_path / 'venv/bin'}/!(activate)"
        in poetry_plugin.get_build_commands()
    )


def test_no_get_rm_command(
    tmp_path, poetry_plugin: plugins.PoetryPlugin, install_path: pathlib.Path
):
    spec = {
        "plugin": "poetry",
        "source": str(tmp_path),
        "poetry-keep-bins": True,
    }
    poetry_plugin._options = plugins.PoetryPluginProperties.unmarshal(spec)
    assert (
        f"rm -rf {install_path / 'venv/bin'}/!(activate)"
        not in poetry_plugin.get_build_commands()
    )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import pathlib

import pytest


@pytest.fixture
def build_path(tmp_path: pathlib.Path) -> pathlib.Path:
    path = tmp_path / "parts" / "foo" / "build"
    path.mkdir(parents=True)
    return path


@pytest.fixture
def install_path(tmp_path: pathlib.Path) -> pathlib.Path:
    path = tmp_path / "parts" / "foo" / "install"
    path.mkdir(parents=True)
    return path



# Copyright 2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for the project service."""

import pathlib
import shutil

import pytest
from craft_application import ServiceFactory

from charmcraft.application.main import APP_METADATA
from charmcraft.services import register_services
from charmcraft.services.project import ProjectService

SAMPLE_PROJECTS_DIR = pathlib.Path(__file__).parent / "sample_projects"


@pytest.fixture
def service_factory(project_path: pathlib.Path) -> ServiceFactory:
    """Override the service factory since we're generating our own project file."""
    register_services()
    factory = ServiceFactory(app=APP_METADATA)
    factory.update_kwargs("project", project_dir=project_path)
    return factory


@pytest.fixture
def service(service_factory: ServiceFactory, project_path):
    return service_factory.get("project")


@pytest.fixture(
    params=[pytest.param(path, id=path.name) for path in SAMPLE_PROJECTS_DIR.iterdir()]
)
def project_data_path(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture
def project_path(project_path: pathlib.Path, project_data_path: pathlib.Path):
    shutil.copytree(project_data_path / "project", project_path, dirs_exist_ok=True)
    return project_path


@pytest.mark.filterwarnings("ignore:The 'charmhub' field")
def test_project_renders_with_build_plan(
    in_project_path: pathlib.Path,
    service: ProjectService,
    service_factory: ServiceFactory,
):
    service.configure(platform=None, build_for=None)

    service.get()

    service_factory.get("build_plan").plan()



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for linters."""

import pathlib
import string
import subprocess
import sys

import pytest
from hypothesis import given, strategies

from charmcraft import linters
from charmcraft.models.lint import LintResult


@given(
    name=strategies.text(
        strategies.characters(
            codec="ascii",  # Only ASCII characters are valid in Python names
            categories=["L", "N"],  # Letters and numbers
            include_characters="._-",  # Or hyphens, underscores, periods
        ),
        min_size=1,  # At least one character
    ).filter(lambda x: x[0] not in "._-" and x[-1] not in "._-"),
    next_char=strategies.sampled_from(string.whitespace + "<>=~!"),
    further_garbage=strategies.text(),
)
def test_fuzz_python_name_regex(name, next_char, further_garbage):
    assert linters.PYTHON_NAME_REGEX.match(name).group(0) == name
    assert linters.PYTHON_NAME_REGEX.match(f"{name}{next_char}").group(0) == name
    assert (
        linters.PYTHON_NAME_REGEX.match(f"{name}{next_char}{further_garbage}").group(0)
        == name
    )


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        (
            r'requests [security,tests] >= 2.8.1, == 2.8.* ; python_version < "2.7"',
            "2.8.1",
        ),
        (">=1.0.0", "1.0.0"),
    ],
)
def test_min_version_regex_matches(string, expected):
    assert linters.MIN_VERSION_REGEX.search(string).group(1) == expected


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        (
            r'requests [security,tests] >= 2.8.1, == 2.8.* ; python_version < "2.7"',
            "2.8",
        ),
        ("~=1.0.0", "1.0"),
    ],
)
def test_approx_version_regex_matches(string, expected):
    assert expected in linters.APPROX_VERSION_REGEX.search(string).group(1, 2)


@pytest.mark.parametrize(
    ("string", "expected"),
    [("==1.0.0", "1.0.0")],
)
def test_exact_version_regex_matches(string, expected):
    assert linters.EXACT_VERSION_REGEX.search(string).group(1) == expected


@pytest.fixture
def valid_venv_path(fake_path) -> pathlib.Path:
    """Create and return a fakefs path that contains a valid venv structure"""
    (fake_path / "venv" / "lib").mkdir(parents=True)
    return fake_path


def test_pip_check_not_venv(fake_path: pathlib.Path):
    lint = linters.PipCheck()
    assert lint.run(fake_path) == LintResult.NONAPPLICABLE
    assert lint.text == "Charm does not contain a Python venv."


def test_pip_invalid_venv(fake_path: pathlib.Path):
    (fake_path / "venv").mkdir()
    lint = linters.PipCheck()
    assert lint.run(fake_path) == LintResult.NONAPPLICABLE
    assert lint.text == "Python venv is not valid."


def test_pip_check_success(valid_venv_path: pathlib.Path, fp):
    fp.register(
        [sys.executable, "-m", "pip", "--python", fp.any(), "check"],
        returncode=0,
        stdout="Loo loo loo, doing pip stuff. Pip stuff is my favourite stuff.",
    )

    lint = linters.PipCheck()
    assert lint.run(valid_venv_path) == LintResult.OK
    assert lint.text == linters.PipCheck.text


def test_pip_check_warning(valid_venv_path: pathlib.Path, fp):
    fp.register(
        [sys.executable, "-m", "pip", "--python", fp.any(), "check"],
        returncode=1,
        stdout="This error was sponsored by Raytheon Knife Missiles™",
    )

    lint = linters.PipCheck()
    assert lint.run(valid_venv_path) == LintResult.WARNING
    assert lint.text == "This error was sponsored by Raytheon Knife Missiles™"


def test_pip_check_exception(valid_venv_path: pathlib.Path, monkeypatch):
    def _raises_eperm(*args, **kwargs) -> None:
        raise PermissionError(13, "Permission denied")

    monkeypatch.setattr(subprocess, "run", _raises_eperm)

    lint = linters.PipCheck()
    assert lint.run(valid_venv_path) == LintResult.NONAPPLICABLE
    assert (
        lint.text
        == f"Permission denied: Could not run Python executable at {sys.executable}."
    )


def test_pip_check_repair_no_bin(valid_venv_path: pathlib.Path, fp):
    """Check that the bin directory is deleted if it was missing before"""
    fp.register(
        [sys.executable, "-m", "pip", "--python", fp.any(), "check"],
        returncode=0,
        stdout="Gosh, I sure hope I remember where everything went.",
    )
    lint = linters.PipCheck()

    # Make sure it doesn't leave behind "bin" if it didn't exist
    assert lint.run(valid_venv_path) == LintResult.OK
    assert lint.text == "Virtual environment is valid."
    assert not (valid_venv_path / "venv" / "bin").exists()


def test_pip_check_repair_no_py(valid_venv_path: pathlib.Path, fp):
    """Check that the python symlink is deleted if it was missing before"""
    fp.register(
        [sys.executable, "-m", "pip", "--python", fp.any(), "check"],
        returncode=0,
        stdout="Gosh, I sure hope I remember where everything went.",
    )
    lint = linters.PipCheck()

    # Make sure it keeps "bin" if only the Python binary didn't exist
    (valid_venv_path / "venv" / "bin").mkdir()
    assert lint.run(valid_venv_path) == LintResult.OK
    assert lint.text == "Virtual environment is valid."
    assert (valid_venv_path / "venv" / "bin").exists()
    assert not (valid_venv_path / "venv" / "bin" / "python").exists()


def test_pip_check_repair_all(valid_venv_path: pathlib.Path, fp):
    """Check that nothing is changed if all components are present"""
    fp.register(
        [sys.executable, "-m", "pip", "--python", fp.any(), "check"],
        returncode=0,
        stdout="Gosh, I sure hope I remember where everything went.",
    )
    lint = linters.PipCheck()

    (valid_venv_path / "venv" / "bin").mkdir()
    (valid_venv_path / "venv" / "bin" / "python").touch()

    assert lint.run(valid_venv_path) == LintResult.OK
    assert lint.text == "Virtual environment is valid."
    assert (valid_venv_path / "venv" / "bin" / "python").is_file()


@pytest.mark.parametrize(
    ("name", "expected"),
    [
        ("cryptography", "cryptography"),
        (
            "opentelemetry.exporter.otlp.proto.http",
            "opentelemetry_exporter_otlp_proto_http",
        ),
        ("poetry-core", "poetry_core"),
    ],
)
def test_pydeps_fs(name, expected):
    assert linters.PyDeps.convert_to_fs(name) == expected


@pytest.mark.parametrize(
    ("spec", "version", "expected"),
    [
        (">=1.0.0", (1, 0, 0), True),
        (">= 2.0.0, == 1.2.3, ==1.2.4, ~=1.4.0", (1, 2, 3), True),
        (">= 2.0.0, == 1.2.3, ==1.2.4, ~=1.4.0", (1, 2, 4), True),
        (">= 2.0.0, == 1.2.3, ==1.2.4, ~=1.4.0, ==1.5.*", (1, 4, 10), True),
        ("==1.5.*", (1, 5, 10), True),
        (">= 2.0.0, == 1.2.3, ==1.2.4, ~=1.4.0", (2, 2, 4), True),
    ],
)
def test_pydeps_version_matches(spec, version, expected):
    assert linters.PyDeps.version_matches(spec, version) == expected


@pytest.mark.parametrize(
    ("deps", "expected"),
    [
        (set(), (set(), set())),
        ({"existing"}, (set(), set())),
        ({"existing.child.package>=1.0.0"}, (set(), set())),
        ({"existing.child.package==1.0.0"}, (set(), {"existing.child.package==1.0.0"})),
        ({"existing-package"}, (set(), set())),
        ({"existing-module"}, (set(), set())),
        ({"nope"}, ({"nope"}, set())),
    ],
)
def test_pydeps_get_missing_deps(tmp_path, mocker, deps, expected):
    packages_path = tmp_path / "lib/python3.14/site-packages"
    (packages_path / "existing").mkdir(parents=True)
    (packages_path / "existing_child_package-1.0.1.dist-info").mkdir(parents=True)
    (packages_path / "existing_package").mkdir(parents=True)
    (packages_path / "existing_module.py").touch()

    assert linters.PyDeps._get_missing_deps(deps, tmp_path) == expected



"""Dynamically load stage-package from another snap into our sys.path.

https://docs.python.org/3/library/site.html#module-sitecustomize
"""

import hashlib
import os
import sys
from pathlib import Path

TARGET_HASH = "8641d59472d54186f68939ab09cd4882cfba8f20efe278e6a539973c0b3e512c84d952e80f1d7297f8fd025e6878ea58"


def _find_pkg() -> None:
    for craft_dir in Path("/snap").glob("*craft*/current"):
        hasher = hashlib.sha3_384()
        hasher.update(str(craft_dir).encode())
        pkg_dir_hash = hasher.hexdigest()
        if pkg_dir_hash != TARGET_HASH:
            continue
        pkg_path = craft_dir / "craft-plugins"
        sys.path.insert(0, os.fspath(pkg_path))
        return


_find_pkg()



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for the Image service."""

import pytest

from charmcraft import application
from charmcraft.services.image import ImageService


@pytest.fixture
def image_service() -> ImageService:
    service = ImageService(
        app=application.APP_METADATA,
        services=None,  # pyright: ignore[reportArgumentType]
    )
    service.setup()
    return service


@pytest.mark.parametrize(
    "url",
    [
        "docker://hello-world@sha256:18a657d0cc1c7d0678a3fbea8b7eb4918bba25968d3e1b0adebfa71caddbc346",
        "hello-world@sha256:18a657d0cc1c7d0678a3fbea8b7eb4918bba25968d3e1b0adebfa71caddbc346",
        "docker://ghcr.io/canonical/charmed-mysql@sha256:89b8305613f6ce94f78a7c9b4baedef78f2816fd6bc74c00f6607bc5e57bd8e6",
        "docker://quay.io/prometheus/blackbox-exporter:v0.24.0",
        "docker://quay.io/prometheus/blackbox-exporter:v0.24.0@sha256:3af31f8bd1ad2907b4b0f7c485fde3de0a8ee0b498d42fc971f0698885c03acb",
    ],
)
def test_get_maybe_id_from_docker_no_exceptions(image_service: ImageService, url):
    image_service.get_maybe_id_from_docker(url)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for the provider service."""

import pathlib
import shutil
import subprocess
import sys

import craft_application
import pytest
from craft_cli.pytest_plugin import RecordingEmitter
from craft_platforms import BuildInfo

from charmcraft.services.provider import _maybe_lock_cache

pytestmark = [pytest.mark.slow]


@pytest.mark.skipif(
    sys.platform == "darwin", reason="multipass sometimes fails weirdly for this test"
)
def test_lock_cache(
    service_factory: craft_application.ServiceFactory,
    tmp_path: pathlib.Path,
    default_build_info: BuildInfo,
    emitter: RecordingEmitter,
):
    cache_path = tmp_path / "cache"
    cache_path.mkdir()
    lock_file = cache_path / "charmcraft.lock"
    bash_lock_cmd = (
        ["bash", "-c", f"flock -n {lock_file} true"] if shutil.which("flock") else None
    )
    provider = service_factory.provider
    provider_kwargs = {
        "build_info": default_build_info,
        "work_dir": tmp_path,
        "cache_path": cache_path,
    }
    assert not lock_file.exists()

    with provider.instance(**provider_kwargs):
        # Test that the cache lock gets created
        assert lock_file.is_file()
        if bash_lock_cmd:
            with pytest.raises(subprocess.CalledProcessError):
                # Another process should not be able to lock the file.
                subprocess.run(bash_lock_cmd, check=True)

    # After exiting we should be able to lock the file.
    if bash_lock_cmd:
        subprocess.run(bash_lock_cmd, check=True)


@pytest.mark.skipif(
    sys.platform == "darwin", reason="multipass sometimes fails weirdly for this test"
)
def test_locked_cache_no_cache(
    service_factory: craft_application.ServiceFactory,
    tmp_path: pathlib.Path,
    default_build_info: BuildInfo,
    emitter: RecordingEmitter,
):
    cache_path = tmp_path / "cache"
    cache_path.mkdir()
    lock_file = cache_path / "charmcraft.lock"

    bash_lock_cmd = (
        ["bash", "-c", f"flock -n {lock_file} true"] if shutil.which("flock") else None
    )
    # Check that we can lock the file from another process.
    if bash_lock_cmd:
        subprocess.run(bash_lock_cmd, check=True)
    _ = _maybe_lock_cache(cache_path)
    # And now we can't.
    if bash_lock_cmd:
        with pytest.raises(subprocess.CalledProcessError):
            subprocess.run(bash_lock_cmd, check=True)

    provider = service_factory.provider
    provider_kwargs = {
        "build_info": default_build_info,
        "work_dir": tmp_path,
        "cache_path": cache_path,
    }

    with provider.instance(**provider_kwargs) as instance:
        # Create a file in the cache and ensure it's not visible in the outer fs
        instance.execute_run(["touch", "/root/.cache/cache_cached"])

        # Because we've already locked the cache, we don't get a subdirectory in
        # the cache, and thus the touch command inside there only affected the
        # instance cache and not the shared cache.
        assert list(cache_path.iterdir()) == [cache_path / "charmcraft.lock"]
        emitter.assert_progress(
            "Shared cache locked by another process; running without cache.",
            permanent=True,
        )

        assert not (tmp_path / "cache_cached").exists()


@pytest.mark.skipif(
    sys.platform == "darwin", reason="multipass sometimes fails weirdly for this test"
)
def test_cache_symlink(
    service_factory: craft_application.ServiceFactory,
    tmp_path: pathlib.Path,
    default_build_info: BuildInfo,
    emitter: RecordingEmitter,
):
    cache_path = tmp_path / "cache"
    cache_path.mkdir()
    provider = service_factory.provider
    provider_kwargs = {
        "build_info": default_build_info,
        "work_dir": tmp_path,
        "cache_path": cache_path,
    }
    with provider.instance(**provider_kwargs) as instance:
        instance.execute_run(["test", "-d", "/root/.cache"], check=True)
        instance.execute_run(
            ["test", "-d", "/root/snap/charmcraft/common/cache"], check=True
        )
        with pytest.raises(subprocess.CalledProcessError):
            instance.execute_run(["test", "-d", "/blorp"], check=True)



# Copyright 2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for the charmcraft-specific project service."""

import pathlib
import re

import craft_platforms
import pytest
import yaml
from craft_cli import CraftError

from charmcraft.application.main import APP_METADATA
from charmcraft.services.project import ProjectService

CURRENT_ARCH = craft_platforms.DebianArchitecture.from_host()


@pytest.fixture
def service(project_path):
    return ProjectService(app=APP_METADATA, services=None, project_dir=project_path)


@pytest.mark.parametrize(
    ("bases", "expected"),
    [
        pytest.param(
            [{"name": "ubuntu", "channel": "20.04"}],
            {
                f"ubuntu-20.04-{CURRENT_ARCH}": {
                    "build-on": [f"ubuntu@20.04:{CURRENT_ARCH}"],
                    "build-for": [f"ubuntu@20.04:{CURRENT_ARCH}"],
                },
            },
            id="focal-current-arch",
        ),
        pytest.param(
            [{"name": "ubuntu", "channel": "22.04", "architectures": ["riscv64"]}],
            {
                "ubuntu-22.04-riscv64": {
                    "build-on": ["ubuntu@22.04:riscv64"],
                    "build-for": ["ubuntu@22.04:riscv64"],
                }
            },
            id="jammy-riscv64-no-cross",
        ),
    ],
)
def test_render_legacy_platforms(project_path, service, bases, expected):
    (project_path / "charmcraft.yaml").write_text(
        yaml.safe_dump({"name": "bases-test", "type": "charm", "bases": bases})
    )

    assert service._app_render_legacy_platforms() == expected


@pytest.mark.parametrize(
    ("bases", "exc_string"),
    [
        (
            [{"name": "ubuntu", "channel": "24.04"}],
            "'ubuntu@24.04'",
        ),
        (
            [
                {"name": "ubuntu", "channel": "22.04"},
                {"name": "ubuntu", "channel": "24.04"},
            ],
            "'ubuntu@24.04'",
        ),
        (
            [
                {"name": "ubuntu", "channel": "25.10"},
                {"name": "ubuntu", "channel": "24.04"},
            ],
            "'ubuntu@24.04', 'ubuntu@25.10'",
        ),
    ],
)
def test_render_legacy_platforms_error(
    project_path: pathlib.Path, service: ProjectService, bases: list[dict], exc_string
):
    (project_path / "charmcraft.yaml").write_text(
        yaml.safe_dump({"name": "bases-test", "type": "charm", "bases": bases})
    )

    with pytest.raises(
        CraftError,
        match=f"^Not valid for use with the 'bases' key: {re.escape(exc_string)}",
    ):
        service._app_render_legacy_platforms()


@pytest.mark.parametrize(
    ("platforms", "expected"),
    [
        *(
            pytest.param(
                {str(arch): None},
                {str(arch): {"build-on": [str(arch)], "build-for": [str(arch)]}},
                id=f"expand-{arch}",
            )
            for arch in craft_platforms.DebianArchitecture
        ),
        *(
            pytest.param(
                {"anything": {"build-on": [str(arch)], "build-for": ["all"]}},
                {"anything": {"build-on": [str(arch)], "build-for": ["all"]}},
                id=f"on-{arch}-for-all",
            )
            for arch in craft_platforms.DebianArchitecture
        ),
        *(
            pytest.param(
                {"unvectored": {"build-on": str(arch), "build-for": "all"}},
                {"unvectored": {"build-on": [str(arch)], "build-for": ["all"]}},
                id=f"vectorise-{arch}",
            )
            for arch in craft_platforms.DebianArchitecture
        ),
        pytest.param(
            {"ppc64el": {"build-on": ["amd64", "riscv64"]}},
            {"ppc64el": {"build-on": ["amd64", "riscv64"], "build-for": ["ppc64el"]}},
            id="only-build-on-valid-name",
        ),
        pytest.param(
            {"all": {"build-on": "riscv64"}},
            {"all": {"build-on": ["riscv64"], "build-for": ["all"]}},
            id="lazy-all",
        ),
        pytest.param(
            {"s390x": {"build-on": "ppc64el", "build-for": None}},
            {"s390x": {"build-on": ["ppc64el"], "build-for": ["s390x"]}},
            id="null-build-for-valid-name",
        ),
        pytest.param(
            {
                "jammy": {
                    "build-on": ["ubuntu@22.04:amd64"],
                    "build-for": ["ubuntu@22.04:amd64"],
                },
                "ubuntu@24.04:riscv64": None,
            },
            {
                "jammy": {
                    "build-on": ["ubuntu@22.04:amd64"],
                    "build-for": ["ubuntu@22.04:amd64"],
                },
                "ubuntu@24.04:riscv64": {
                    "build-on": ["ubuntu@24.04:riscv64"],
                    "build-for": ["ubuntu@24.04:riscv64"],
                },
            },
            id="multi-platform",
        ),
    ],
)
def test_get_platforms_correct(
    project_path: pathlib.Path, service: ProjectService, platforms, expected
):
    (project_path / "charmcraft.yaml").write_text(
        yaml.safe_dump(
            {
                "name": "platforms-test",
                "type": "charm",
                "summary": "a summary",
                "description": "a description",
                "parts": {"something": {"plugin": "nil"}},
                "platforms": platforms,
            }
        )
    )

    assert service.get_platforms() == expected
    # Check that it renders.
    service.configure(platform=None, build_for=None)
    service.get()
    assert service.get().marshal()["platforms"] == expected



# Copyright 2023,2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for project-related models."""

import itertools
import json
import pathlib
import textwrap
from textwrap import dedent
from typing import Any

import craft_cli.pytest_plugin
import hypothesis
import pydantic
import pyfakefs.fake_filesystem
import pytest
from craft_application import util
from craft_application.errors import CraftValidationError
from craft_application.util import safe_yaml_load
from craft_cli import CraftError
from craft_providers import bases
from hypothesis import strategies

from charmcraft import const
from charmcraft.models import project
from charmcraft.models.charmcraft import Base

SIMPLE_BASE = Base(name="simple", channel="0.0")
BASE_WITH_ONE_ARCH = Base(name="arch", channel="1.0", architectures=["amd64"])
BASE_WITH_MULTIARCH = Base(
    name="multiarch", channel="2.0", architectures=["arm64", "riscv64"]
)
SIMPLE_BASENAME = bases.BaseName("simple", "0.0")
ONE_ARCH_BASENAME = bases.BaseName("arch", "1.0")
MULTIARCH_BASENAME = bases.BaseName("multiarch", "2.0")
UBUNTU_JAMMY = Base(name="ubuntu", channel="22.04", architectures=["amd64"])
SIMPLE_BASES = (
    UBUNTU_JAMMY,
    Base(name="almalinux", channel="9", architectures=["amd64"]),
)
COMPLEX_BASES = (
    Base(
        name="ubuntu",
        channel="devel",
        architectures=["amd64", "arm64", "s390x", "riscv64"],
    ),
    Base(
        name="almalinux",
        channel="9",
        architectures=["amd64", "arm64", "s390x", "ppc64el"],
    ),
)
SIMPLE_BASE_CONFIG_DICT = {"name": "ubuntu", "channel": "22.04"}
FULL_BASE_CONFIG_DICT = {
    "build-on": [{"channel": "22.04", "name": "ubuntu"}],
    "run-on": [{"channel": "22.04", "name": "ubuntu"}],
}
BASIC_CHARM_PARTS = {"charm": {"plugin": "charm", "source": "."}}
BASIC_CHARM_PARTS_EXPANDED = {
    "charm": {
        "plugin": "charm",
        "source": ".",
        "charm-binary-python-packages": [],
        "charm-entrypoint": "src/charm.py",
        "charm-python-packages": [],
        "charm-requirements": [],
        "charm-strict-dependencies": False,
    }
}

MINIMAL_CHARMCRAFT_YAML = f"""\
type: charm
bases:
  - build-on:
      - name: ubuntu
        channel: "22.04"
        architectures: [{util.get_host_architecture()}]
    run-on:
      - name: ubuntu
        channel: "22.04"
        architectures: [arm64]
"""
MINIMAL_CHARMCRAFT_DICT = {
    "type": "charm",
    "bases": [
        {
            "build-on": [
                {
                    "name": "ubuntu",
                    "channel": "22.04",
                    "architectures": [util.get_host_architecture()],
                },
            ],
            "run-on": [
                {"name": "ubuntu", "channel": "22.04", "architectures": ["arm64"]},
            ],
        }
    ],
}
SIMPLE_METADATA_YAML = (
    "{name: charmy-mccharmface, summary: Charmy!, description: Very charming!}"
)
SIMPLE_CHARMCRAFT_YAML = f"""\
type: charm
name: charmy-mccharmface
summary: Charmy!
description: Very charming!
bases:
  - build-on:
      - name: ubuntu
        channel: "22.04"
        architectures: [{util.get_host_architecture()}]
    run-on:
      - name: ubuntu
        channel: "22.04"
        architectures: [arm64]
"""
SIMPLE_CHARMCRAFT_DICT = MINIMAL_CHARMCRAFT_DICT | {
    "name": "charmy-mccharmface",
    "summary": "Charmy!",
    "description": "Very charming!",
}
SIMPLE_CONFIG_YAML = (
    "options: {admin: {default: root, description: Admin user, type: string}}"
)
SIMPLE_CONFIG_DICT = {
    "options": {
        "admin": {"type": "string", "default": "root", "description": "Admin user"}
    }
}
SIMPLE_ACTIONS_YAML = "snooze: {description: Take a little nap.}"
SIMPLE_ACTIONS_DICT = {"snooze": {"description": "Take a little nap."}}
CHARMCRAFT_YAML_NON_VECTORISED_PLATFORMS = """\
type: charm
name: test-1874-regression
summary: Regression test for #1874
description: A charm for regression testing https://github.com/canonical/charmcraft/issues/1874

parts:
  charm:
    plugin: dump
    source: .
    prime:
      - actions/*
      - files/*
      - hooks/*
      - lib/*
      - templates/*
      - config.yaml
      - copyright
      - icon.svg
      - LICENSE
      - Makefile
      - metadata.yaml
      - README.md

base: ubuntu@24.04
platforms:
 amd64:
   build-on: amd64
   build-for: amd64
 arm64:
   build-on: arm64
   build-for: arm64
 ppc64el:
   build-on: ppc64el
   build-for: ppc64el
 s390x:
   build-on: s390x
   build-for: s390x
"""


# region CharmPlatform tests
@pytest.mark.parametrize(
    ("bases", "expected"),
    [
        (
            [Base(name="xname", channel="xchannel", architectures=["xarch1"])],
            "xname-xchannel-xarch1",
        ),
        (
            [
                Base(
                    name="xname",
                    channel="xchannel",
                    architectures=["xarch1", "xarch2", "xarch3"],
                )
            ],
            "xname-xchannel-xarch1-xarch2-xarch3",
        ),
        (
            [
                Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
                Base(
                    name="x2name",
                    channel="x2channel",
                    architectures=["x2arch1", "x2arch2"],
                ),
            ],
            "x1name-x1channel-x1arch_x2name-x2channel-x2arch1-x2arch2",
        ),
    ],
)
def test_platform_from_bases_backwards_compatible(bases: list[Base], expected: str):
    """Replicates the format_charm_file_name tests in test_package.py.

    This ensures that charm names remain consistent as we move to platforms.
    """
    assert project.get_charm_file_platform_str(bases) == expected


@pytest.mark.parametrize("base", [*SIMPLE_BASES, *COMPLEX_BASES])
def test_platform_from_single_base(base):
    """Tests generating a platform name from a real (single) base."""
    expected_architectures = "-".join(base.architectures)
    expected = f"{base.name}-{base.channel}-{expected_architectures}"

    actual = project.get_charm_file_platform_str([base])

    assert actual == expected


@pytest.mark.parametrize(
    ("bases", "expected"),
    [
        (SIMPLE_BASES, "ubuntu-22.04-amd64_almalinux-9-amd64"),
        (
            COMPLEX_BASES,
            "ubuntu-devel-amd64-arm64-s390x-riscv64_almalinux-9-amd64-arm64-s390x-ppc64el",
        ),
    ],
)
def test_platform_from_multiple_bases(bases, expected):
    assert project.get_charm_file_platform_str(bases) == expected


VALID_PLATFORM_ARCHITECTURES = [
    *(
        list(x) for x in itertools.combinations(const.CharmArch, 1)
    ),  # A single architecture in a list
    *(
        list(x) for x in itertools.combinations(const.CharmArch, 2)
    ),  # Two architectures in a list
]


@pytest.mark.parametrize(
    ("lib_name", "expected_lib_name"),
    [
        ("charm.lib", "charm.lib"),
        ("charm_with_underscores.lib", "charm-with-underscores.lib"),
        ("charm-with-hyphens.lib", "charm-with-hyphens.lib"),
        ("charm.lib_with_hyphens", "charm.lib_with_hyphens"),
        ("charm0.number_0_lib", "charm0.number_0_lib"),
    ],
)
@pytest.mark.parametrize("lib_version", ["0", "1", "2.0", "2.1", "3.14"])
def test_create_valid_charm_lib(
    lib_name: str, expected_lib_name: str, lib_version: str
):
    lib = project.CharmLib.unmarshal({"lib": lib_name, "version": lib_version})
    assert lib.lib == expected_lib_name


@pytest.mark.parametrize(
    ("name", "error_match"),
    [
        (
            "boop",
            r"Library name invalid. Expected '\[charm_name\].\[lib_name\]', got 'boop'",
        ),
        (
            "Invalid charm name.valid_lib",
            "Invalid charm name for lib 'Invalid charm name.valid_lib'. Value 'Invalid charm name' is invalid",
        ),
        (
            "my_charm.invalid-library-name",
            "Library name 'invalid-library-name' is invalid. Library names must be valid Python module names.",
        ),
    ],
)
def test_invalid_charm_lib_name(name: str, error_match: str):
    with pytest.raises(pydantic.ValidationError, match=error_match):
        project.CharmLib.unmarshal({"lib": name, "version": "0"})


@hypothesis.given(
    strategies.one_of(
        strategies.floats(
            min_value=0.001,
            max_value=2**32,
            allow_nan=False,
            allow_infinity=False,
            allow_subnormal=False,
        ),
        strategies.integers(min_value=0, max_value=2**32),
    )
)
def test_valid_library_version(version: float):
    project.CharmLib.unmarshal({"lib": "charm_name.lib_name", "version": str(version)})


@pytest.mark.parametrize("version", [".1", "NaN", ""])
def test_invalid_api_version(version: str):
    with pytest.raises(
        pydantic.ValidationError,
        match="API version not valid. Expected an integer, got '",
    ):
        project.CharmLib(lib="charm_name.lib_name", version=version)


@pytest.mark.parametrize("version", ["1.", "1.number"])
def test_invalid_patch_version(version: str):
    with pytest.raises(
        pydantic.ValidationError,
        match="Patch version not valid. Expected an integer, got '",
    ):
        project.CharmLib(lib="charm_name.lib_name", version=version)


@pytest.mark.parametrize("version", [1, 1.1])
def test_bad_version_type(version: Any):
    with pytest.raises(
        pydantic.ValidationError,
        match="1 validation error for CharmLib\nversion\n.+Input should be a valid string",
    ):
        project.CharmLib(lib="charm_name.lib_name", version=version)


# endregion
# region CharmcraftProject tests
@pytest.mark.parametrize(
    ("data", "type_class"),
    [
        (
            {
                "type": "charm",
                "name": "basic",
                "summary": "A basic charm",
                "description": "A thing",
                "bases": [SIMPLE_BASE_CONFIG_DICT],
            },
            project.Charm,
        ),
    ],
)
def test_unmarshal_success(data, type_class):
    assert isinstance(project.CharmcraftProject.unmarshal(data), type_class)


def test_unmarshal_bundle():
    with pytest.raises(CraftValidationError, match="Invalid 'charmcraft.yaml' file"):
        project.CharmcraftProject.unmarshal({"type": "bundle"})


@pytest.mark.parametrize("type_", [None, "", "invalid", "Dvorak"])
def test_unmarshal_invalid_type(type_):
    with pytest.raises(ValueError, match="^field type cannot be "):
        project.CharmcraftProject.unmarshal({"type": type_})


@pytest.mark.parametrize(
    (
        "charmcraft_yaml",
        "metadata_yaml",
        "config_yaml",
        "actions_yaml",
        "expected_dict",
    ),
    [
        (
            SIMPLE_CHARMCRAFT_YAML,
            None,
            None,
            None,
            SIMPLE_CHARMCRAFT_DICT | {"parts": BASIC_CHARM_PARTS},
        ),
        (
            MINIMAL_CHARMCRAFT_YAML,
            SIMPLE_METADATA_YAML,
            None,
            None,
            SIMPLE_CHARMCRAFT_DICT | {"parts": BASIC_CHARM_PARTS},
        ),
        (
            SIMPLE_CHARMCRAFT_YAML,
            None,
            SIMPLE_CONFIG_YAML,
            None,
            SIMPLE_CHARMCRAFT_DICT
            | {"config": SIMPLE_CONFIG_DICT, "parts": BASIC_CHARM_PARTS},
        ),
        (
            SIMPLE_CHARMCRAFT_YAML,
            None,
            None,
            SIMPLE_ACTIONS_YAML,
            SIMPLE_CHARMCRAFT_DICT
            | {"actions": SIMPLE_ACTIONS_DICT, "parts": BASIC_CHARM_PARTS},
        ),
        (
            MINIMAL_CHARMCRAFT_YAML,
            SIMPLE_METADATA_YAML,
            SIMPLE_CONFIG_YAML,
            SIMPLE_ACTIONS_YAML,
            SIMPLE_CHARMCRAFT_DICT
            | {
                "actions": SIMPLE_ACTIONS_DICT,
                "config": SIMPLE_CONFIG_DICT,
                "parts": BASIC_CHARM_PARTS,
            },
        ),
        pytest.param(
            SIMPLE_CHARMCRAFT_YAML
            + textwrap.dedent(
                """\
                parts:
                  charm: {}
                  reactive: {}
                """
            ),
            None,
            None,
            None,
            SIMPLE_CHARMCRAFT_DICT
            | {
                "parts": {
                    "charm": {
                        "plugin": "charm",
                        "source": ".",
                    },
                    "reactive": {
                        "plugin": "reactive",
                    },
                }
            },
            id="implicit-parts-plugins",
        ),
        pytest.param(
            CHARMCRAFT_YAML_NON_VECTORISED_PLATFORMS,
            None,
            None,
            None,
            {
                "name": "test-1874-regression",
                "summary": "Regression test for",
                "description": "A charm for regression testing https://github.com/canonical/charmcraft/issues/1874",
                "base": "ubuntu@24.04",
                "platforms": {
                    "amd64": {"build-on": ["amd64"], "build-for": ["amd64"]},
                    "arm64": {"build-on": ["arm64"], "build-for": ["arm64"]},
                    "ppc64el": {"build-on": ["ppc64el"], "build-for": ["ppc64el"]},
                    "s390x": {"build-on": ["s390x"], "build-for": ["s390x"]},
                },
                "parts": {
                    "charm": {
                        "plugin": "dump",
                        "source": ".",
                        "prime": [
                            "actions/*",
                            "files/*",
                            "hooks/*",
                            "lib/*",
                            "templates/*",
                            "config.yaml",
                            "copyright",
                            "icon.svg",
                            "LICENSE",
                            "Makefile",
                            "metadata.yaml",
                            "README.md",
                        ],
                    }
                },
                "type": "charm",
            },
            id="1874-regression",
        ),
    ],
)
def test_from_yaml_file_success(
    fs: pyfakefs.fake_filesystem.FakeFilesystem,
    simple_charm,
    charmcraft_yaml: str,
    metadata_yaml: str | None,
    config_yaml: str | None,
    actions_yaml: str | None,
    expected_dict: dict[str, Any],
):
    fs.create_file("/charmcraft.yaml", contents=charmcraft_yaml)
    if metadata_yaml:
        fs.create_file("/metadata.yaml", contents=metadata_yaml)
    if config_yaml:
        fs.create_file("/config.yaml", contents=config_yaml)
    if actions_yaml:
        fs.create_file("/actions.yaml", contents=actions_yaml)

    actual = project.CharmcraftProject.from_yaml_file(pathlib.Path("/charmcraft.yaml"))

    assert actual.marshal() == expected_dict


@pytest.mark.parametrize(
    (
        "charmcraft_yaml",
        "metadata_yaml",
        "config_yaml",
        "actions_yaml",
        "exc_class",
        "match",
        "details",
    ),
    [
        pytest.param(
            None,
            None,
            None,
            None,
            CraftError,
            r"^Could not find charmcraft\.yaml at '.charmcraft\.yaml'$",
            None,
            id="FileNotFound",
        ),
        pytest.param(
            f"{SIMPLE_CHARMCRAFT_YAML}\nconfig: ",
            None,
            SIMPLE_CONFIG_YAML,
            None,
            CraftValidationError,
            r"^Cannot specify 'config' section",
            None,
            id="duplicate-config",
        ),
        pytest.param(
            f"{SIMPLE_CHARMCRAFT_YAML}\nactions:",
            None,
            None,
            SIMPLE_ACTIONS_YAML,
            CraftValidationError,
            r"^Cannot specify 'actions' section",
            None,
            id="duplcate-actions",
        ),
    ],
)
def test_from_yaml_file_exception(
    fs: pyfakefs.fake_filesystem.FakeFilesystem,
    charmcraft_yaml: str | None,
    metadata_yaml: str | None,
    config_yaml: str | None,
    actions_yaml: str | None,
    exc_class: type[CraftError],
    match: str,
    details: str,
):
    if charmcraft_yaml:
        fs.create_file("/charmcraft.yaml", contents=charmcraft_yaml)
    if metadata_yaml:
        fs.create_file("/metadata.yaml", contents=metadata_yaml)
    if config_yaml:
        fs.create_file("/config.yaml", contents=config_yaml)
    if actions_yaml:
        fs.create_file("/actions.yaml", contents=actions_yaml)

    with pytest.raises(exc_class, match=match) as exc:
        project.CharmcraftProject.from_yaml_file(pathlib.Path("/charmcraft.yaml"))

    assert exc.value.details == details


@pytest.mark.parametrize(
    ("cls", "content"),
    [
        (
            project.BasesCharm,
            {
                "type": "charm",
                "name": "blah",
                "summary": "",
                "description": "",
                "bases": [{"name": "ubuntu", "channel": "22.04"}],
                "charmhub": {"api_url": "http://charmhub.io"},
            },
        ),
        (
            project.PlatformCharm,
            {
                "type": "charm",
                "name": "blah",
                "summary": "",
                "description": "",
                "base": "ubuntu@24.04",
                "platforms": {"amd64": None},
                "parts": {"my-part": {"plugin": "nil"}},
                "charmhub": {"api_url": "http://charmhub.io"},
            },
        ),
    ],
)
def test_warn_on_deprecated_charmhub(
    emitter: craft_cli.pytest_plugin.RecordingEmitter, cls, content
):
    with pytest.warns(DeprecationWarning, match="'charmhub' field"):
        cls.model_validate(content)
    emitter.assert_progress(
        "WARNING: The 'charmhub' field is deprecated and no longer used. It will be removed in a "
        f"future release. Use the ${const.STORE_API_ENV_VAR}, ${const.STORE_STORAGE_ENV_VAR} and "
        f"${const.STORE_REGISTRY_ENV_VAR} environment variables instead.",
        permanent=True,
    )


# endregion
# region Charm tests
@pytest.mark.parametrize(
    ("values", "expected_changes"),
    [
        pytest.param(
            {"bases": [SIMPLE_BASE_CONFIG_DICT]},
            {"bases": [FULL_BASE_CONFIG_DICT]},
            id="simple-base",
        ),
    ],
)
def test_instantiate_bases_charm_success(
    values: dict[str, Any], expected_changes: dict[str, Any]
):
    """Various successful instantiations of a charm project."""
    values.update(
        {
            "type": "charm",
            "name": "test-charm",
            "summary": "A test charm",
            "description": "A test charm model that can be successfully instantiated.",
        }
    )
    expected = values.copy()
    expected.update(expected_changes)

    actual = project.BasesCharm.model_validate(values)

    assert actual.marshal() == expected


@pytest.mark.parametrize(
    ("values", "error_cls", "error_match"),
    [
        pytest.param(
            {
                "type": "charm",
                "name": "test-charm",
                "summary": "A test charm",
                "description": "This charm has no bases and is thus invalid.",
            },
            pydantic.ValidationError,
            r"bases\s+Field required",
            id="no-bases",
        ),
        pytest.param(
            {
                "type": "charm",
                "name": "test-charm",
                "summary": "A test charm",
                "description": "Empty bases, also invalid.",
                "bases": [],
            },
            pydantic.ValidationError,
            r"bases\s+List should have at least 1 item",
            id="empty-bases",
        ),
    ],
)
def test_instantiate_bases_charm_error(
    values: dict[str, Any], error_cls: type[Exception], error_match: str
):
    with pytest.raises(error_cls, match=error_match):
        project.BasesCharm(**values)


@pytest.mark.parametrize("base", ["ubuntu@18.04", "ubuntu@22.04"])
def test_devel_bases(monkeypatch, base):
    monkeypatch.setattr(const, "DEVEL_BASE_STRINGS", [base])

    with pytest.raises(
        pydantic.ValidationError,
        match=r"requires a build-base \(recommended: 'build-base: ubuntu@devel'\)",
    ):
        project.PlatformCharm(
            type="charm",
            name="test-charm",
            summary="",
            description="",
            base=base,
            platforms={"amd64": None},
            parts={"charm": {"plugin": "charm"}},
        )


@pytest.mark.parametrize(
    "filename",
    [f.name for f in (pathlib.Path(__file__).parent / "valid_charms_yaml").iterdir()],
)
def test_read_charm_from_yaml_file_self_contained_success(tmp_path, filename: str):
    file_path = pathlib.Path(__file__).parent / "valid_charms_yaml" / filename
    with file_path.open() as f:
        expected_dict = safe_yaml_load(f)

    charm = project.CharmcraftProject.from_yaml_file(file_path)

    # The JSON round-trip here is to get rid of any Pydantic constraint objects like AnyHttpUrl
    assert json.loads(json.dumps(charm.marshal())) == expected_dict


@pytest.mark.parametrize(
    ("filename", "errors"),
    [
        (
            "basic.yaml",
            dedent(
                """\
                Bad basic.yaml content:
                - field 'name' required in top-level configuration
                - field 'summary' required in top-level configuration
                - field 'description' required in top-level configuration
                - field 'platforms' required in top-level configuration
                - field 'parts' required in top-level configuration"""
            ),
        ),
        (
            "invalid-type.yaml",
            dedent(
                """\
                Bad charmcraft.yaml content:
                - field type cannot be 'invalid'"""
            ),
        ),
        (
            "invalid-base.yaml",
            dedent(
                """\
                Bad invalid-base.yaml content:
                - base requires 'platforms' definition: {'name': 'ubuntu', 'channel': '24.04'} (in field 'bases[0]')
                - base requires 'platforms' definition: {'name': 'ubuntu', 'channel': 'devel'} (in field 'bases[1]')"""
            ),
        ),
        pytest.param(
            "platforms-no-parts.yaml",
            dedent(
                """\
                Bad platforms-no-parts.yaml content:
                - field 'parts' required in top-level configuration"""
            ),
            id="no-parts-in-platform-charm",
        ),
        pytest.param(
            "platforms-empty-parts.yaml",
            dedent(
                """\
                Bad platforms-empty-parts.yaml content:
                - dictionary should have at least 1 item after validation, not 0 (in field 'parts')"""
            ),
            id="empty-parts-in-platform-charm",
        ),
    ],
)
def test_read_charm_from_yaml_file_error(filename, errors):
    file_path = pathlib.Path(__file__).parent / "invalid_charms_yaml" / filename

    with pytest.raises(CraftValidationError) as exc:
        _ = project.CharmcraftProject.from_yaml_file(file_path)

    assert exc.value.args[0] == errors


@pytest.mark.parametrize(
    ("base", "expected"),
    [
        ({"name": "ubuntu", "channel": "18.04"}, True),
        ({"name": "ubuntu", "channel": "20.04"}, True),
        ({"name": "ubuntu", "channel": "22.04"}, True),
        ({"name": "ubuntu", "channel": "24.04"}, False),
        ({"name": "ubuntu", "channel": "24.10"}, False),
        ({"name": "ubuntu", "channel": "25.04"}, False),
        ({"name": "almalinux", "channel": "9"}, True),
    ],
)
def test_check_legacy_bases(base, expected):
    assert project._check_base_is_legacy(base) == expected


# endregion
